#!/usr/local/bin/gosh

;; For Emacsen (evaluate with C-x C-e) :
;; (outline-minor-mode 1)
;; (setq-local outline-regexp "^[\s\t]*;;[\s]+\\+[+-]*\s")
;; (setq-local outline-level (lambda () (- (outline-level) 4)))

(add-load-path ".")

(use srfi-1)     ;; list utils
(use srfi-13)    ;; string utils
(use util.match) ;; match / match-lambda
(use lmn.peg)
(use lmn.util)

;; + utilities
;;   + test

(define (set-equal? lst1 lst2)
  ;;  Check that two lists are equal as set.
  (and (every (cut member <> lst1) lst2)
       (every (cut member <> lst2) lst1)))

(define-macro (define-test name :rest body)
  ;;  Do a simple test named NAME.
  ;;  BODY is evaluated in sequence like "begin", but when a sequence
  ;; like 'EXP1 ==> EXP2' is found in the BODY, then (equal? EXP1
  ;; EXP2) is asserted and evaluated to 'EXP1'. Similarly 'EXP -->
  ;; LST' asserts (set-equal? EXP LST), 'EXP :: PRED' asserts (PRED
  ;; EXP), 'EXP <: LST' asserts (member EXP LST) respectively.
  (define (process-body list)
    (cond [(not (list? list)) list]
          [else (set! list (map process-body list))
                (cond [(< (length list) 2) list]
                      [(eq? (cadr list) '::)
                       (cons `(let1 _val_ ,(car list)
                                (unless (,(caddr list) _val_)
                                  (_fail ,(x->string (car list))
                                         (x->string _val_)))
                                _val_)
                             (process-body (cdddr list)))]
                      [(eq? (cadr list) '==>)
                       (cons `(rlet1 _val_ ,(car list)
                                (unless (equal? ,(caddr list) _val_)
                                  (_fail ,(x->string (car list))
                                         (x->string _val_))))
                             (process-body (cdddr list)))]
                      [(eq? (cadr list) '-->)
                       (cons `(rlet1 _val_ ,(car list)
                                (unless (set-equal? ,(caddr list) _val_)
                                  (_fail ,(x->string (car list))
                                         (x->string _val_))))
                             (process-body (cdddr list)))]
                      [(eq? (cadr list) '::)
                       (cons `(rlet1 _val_ ,(car list)
                                (unless (,(caddr list) _val_)
                                  (_fail ,(x->string (car list))
                                         (x->string _val_))))
                             (process-body (cdddr list)))]
                      [(eq? (cadr list) '<:)
                       (cons `(rlet1 _val_ ,(car list)
                                (unless (member _val_ ,(caddr list))
                                  (_fail ,(x->string (car list))
                                         (x->string _val_))))
                             (process-body (cdddr list)))]
                      [else
                       (cons (car list)
                             (process-body (cdr list)))])]))
  `(let* ([_name ,name]
          [_fail (^(a b) (error #`"test \",|_name|\" failed.\n\t,|a| = ,|b|"))])
     ,@(process-body body)
     (print #`"test \",|_name|\" passed.")))

(define true? identity)

;;   + exception

;; exception that can be handled within LMNtal REPL
(define-condition-type <lmn-exception> <error> #f)

(define (lmn-error :rest args)
  ;;  Throw a lmn-error.
  (apply error <lmn-exception> args))

;; + atoms
;;   + portptrs

;; an <portptr> is a pointer to a port of atom.
(define-class <portptr> ()
  ((atom :init-keyword :atom) ;; => atom which the ptr is pointing
   (nth :init-keyword :nth))) ;; => index of the port which the ptr is pointing

;; (ptr 'atom) => atom which PTR is pointing
;; (ptr 'nth) => index of the port which PTR is pointing
(define-method object-apply ((ptr <portptr>) (ref <symbol>))
  (case ref
    [(atom) (slot-ref ptr 'atom)]
    [(nth) (slot-ref ptr 'nth)]
    [else (error #`"portptr: invalid slot ,|ref|.")]))

;; two ptrs are equal if and only if they points exactly the same
;; port of atom
(define-method object-equal? ((ptr1 <portptr>) (ptr2 <portptr>))
  (and (eq? (ptr1 'atom) (ptr2 'atom))
       (= (ptr1 'nth) (ptr2 'nth))))

;;   + atoms

;; an <atom> is a pair of name and argument(s)
(define-class <atom> ()
  ((name :init-keyword :name)   ;; => a string
   (args :init-keyword :args))) ;; => a vector of portptrs

;; two atoms are equal if and only if they are exactly the same object
(define-method object-equal? ((atom1 <atom>) (atom2 <atom>))
  (eq? atom1 atom2))

;; (atom 0) ==> 1st argument of atom as a <portptr>
(define-method object-apply ((atom <atom>) (nth <integer>))
  (vector-ref (slot-ref atom 'args) nth))
(define-method (setter object-apply) ((atom <atom>) (nth <integer>) ptr)
  (vector-set! (slot-ref atom 'args) nth ptr))

;; (atom '(1)) ==> generate a <portptr> the 2nd port of ATOM
(define-method object-apply ((atom <atom>) (nth <pair>))
  (make <portptr> :atom atom :nth (car nth)))

;; (atom 'name) ==> name of ATOM
;; (atom 'arity) ==> arity of ATOM
;; (atom 'functor) ==> functor of ATOM
;; (atom 'args) ==> vector of arguments of ATOM (for internal use)
(define-method object-apply ((atom <atom>) (ref <symbol>))
  (case ref
    [(name) (slot-ref atom 'name)]
    [(arity) (vector-length (slot-ref atom 'args))]
    [(functor) #`",(slot-ref atom 'name)_,(vector-length (slot-ref atom 'args))"]
    [(args) (slot-ref atom 'args)]
    [else (error #`"atom: invalid slot ,|ref|.")]))

;;   + basic operations

(define (make-atom name arity)
  ;;  Allocate and initialize an ARITY-ary atom named NAME
  (make <atom> :name name :args (make-vector arity)))

(define (atom-copy atom)
  ;; Make an atom that has the same properties as ATOM. All
  ;; self-pointing links are resolved, but outgoing links are just
  ;; copied. This may produce a ill-formed process.
  (rlet1 newatom (make <atom>
                   :name (atom 'name)
                   :args (vector-copy (atom 'args)))
    (dotimes (n (newatom 'arity))
      (let1 oldarg (atom n)
        (when (and (not (undefined? oldarg))
                   (eq? (oldarg 'atom) atom))
          (set! (newatom n) (newatom (list (oldarg 'nth)))))))))

(define (partner ptr)
  ;; Get a <portptr> which the port this ptr is pointing is
  ;; pointing. For well-formed processes, the container of this ptr is
  ;; returned as a <portptr>.
  ((ptr 'atom) (ptr 'nth)))

;;   + tests / examples

(define-test "basic methods of atoms / portptrs"
  (let ([atom1 (make-atom "foo" 1)]
        [atom2 (make-atom "foo" 1)]
        [atom3 (make-atom "bar" 1)])
    (atom1 'name) ==> "foo"
    (atom1 'arity) ==> 1
    (atom1 'functor) ==> "foo_1"
    (atom1 0) :: undefined?
    (equal? atom1 atom1) :: true?
    (equal? atom1 atom2) ==> #f
    ;; let atom1's 1st argument point atom2's 1nd port
    (set! (atom1 0) (atom2 '(0)))
    ;; let atom2's 1st argument point atom2's 1nd port
    (set! (atom2 0) (atom1 '(0)))
    ;; now they're connected together
    ((partner (atom1 0)) 'atom) ==> atom1
    (let1 atom1! (atom-copy atom1)
      (atom1! 'functor) ==> "foo_1"
      ;; connect atom1! and atom3
      (set! (atom1! 0) (atom3 '(0)))
      (set! (atom3 0) (atom1! '(0)))
      ((atom1! 0) 'atom) ==> atom3
      ;; atom1 is still conencted to atom2
      ((atom1 0) 'atom) ==> atom2)))

;; + processes
;;   + processes

(define-class <process> ()
  ((atoms :init-keyword :atoms
          :init-form (make-hash-table 'string=?)) ;; => hashtable of (functor -> atoms)
   (proxy :init-keyword :proxy))) ;; => proxy atom that handles incoming links

;; (process 'arity) ==> number of freelinks in PROCESS
;; (process 'proxy) ==> proxy atom of PROCESS (for internal use)
;; (process 'atoms) ==> hash-table of lists of atoms (for internal use)
(define-method object-apply ((process <process>) (ref <symbol>))
  (case ref
    [(arity) ((slot-ref process 'proxy) 'arity)]
    [(proxy) (slot-ref process 'proxy)]
    [(atoms) (slot-ref process 'atoms)]
    [else (error #`"process: invalid slot ,|ref|.")]))

;; (process '(1)) ==> generate a <portptr> which points the 2nd port
(define-method object-apply ((process <process>) (nth <pair>))
  ((process 'proxy) (car nth)))
(define-method (setter object-apply) ((process <process>) (nth <pair>) ptr)
  (set! ((process 'proxy) (car nth)) ptr))

;; (process 1) ==> 2nd argument
(define-method object-apply ((process <process>) (nth <integer>))
  (let1 port ((process 'proxy) nth)
    (when (undefined? port)
      (error #`"process: reference to unassigned free port ,|nth|."))
    ((port 'atom) (port 'nth)))) ;; = (partner port)
(define-method (setter object-apply) ((process <process>) (nth <integer>) ptr)
  (let1 port ((process 'proxy) nth)
    (when (undefined? port)
      (error #`"process: reference to unassigned free port ,|nth|."))
    (set! ((port 'atom) (port 'nth)) ptr)))

;; (process "y_1") ==> list of all atoms with functor "y_1" in PROCESS
(define-method object-apply ((process <process>) (functor <string>))
  (hash-table-get (process 'atoms) functor ()))
(define-method (setter object-apply) ((process <process>) (functor <string>) list)
  (if (null? list)
      (hash-table-delete! (process 'atoms) functor)
      (hash-table-put! (process 'atoms) functor list)))

;;   + basic operations

(define (make-process :optional (arity 0))
  ;; Make an empty ARITY-ary process
  (make <process> :proxy (make-atom "" arity)))

(define (process-push! process :rest atoms)
  ;; Push ATOMS to PROCESS. Note that ports are not managed.
  (dolist (atom atoms)
    (let* ([functor (atom 'functor)]
           [lst (process functor)])
      (unless (memq atom lst)
        (set! (process functor) (cons atom lst))))))

(define (process-get process :optional functor)
  ;; Get an atom from PROCESS. If there're no atoms, return #f. When
  ;; optional arg FUNCTOR is specified, get an atom whose functor is
  ;; FUNCTOR.
  (cond [(string? functor)
         (let1 lst (process functor)
           (and (pair? lst) (car lst)))]
        [(undefined? functor)
         (let1 atoms
             ;; get an arbitrary list of atoms
             (let/cc ret (hash-table-for-each (process 'atoms) (^(k v) (ret v))))
           (and (pair? atoms) (car atoms)))]
        [else (error #`"process-get: invalid criteria ,|functor|.")]))

(define (process-pop! process :optional criteria)
  ;; Pop an atom from process and return it. If there're no atoms to
  ;; pop, return #f. When optional arg CRITERIA is a functor, popped
  ;; atom must have the functor. When CRITERIA is an atom, just remove
  ;; it from PROCESS. Note that ports are not managed.
  (cond [(string? criteria)
         (let1 lst (process criteria)
           (and (pair? lst)
                (begin (set! (process criteria) (cdr lst))
                       (car lst))))]
        [(is-a? criteria <atom>)
         (let* ([functor (criteria 'functor)]
                [lst (process functor)])
           (and (memq criteria lst)
                (begin (set! (process functor) (delete1! criteria lst))
                       criteria)))]
        [(undefined? criteria)
         (let1 atoms
             (let/cc ret (hash-table-for-each (process 'atoms) (^(k v) (ret v))))
           (and (pair? atoms)
                (begin (set! (process ((car atoms) 'functor)) (cdr atoms))
                       (car atoms))))]
        [else (error #`"process-pop!: invalid criteria ,|criteria|.")]))

(define (process-member process atom)
  ;; Check if PROCESS has ATOM.
  (memq atom (process (atom 'functor))))

(define (process-copy process)
  ;; Make a process with the same member as PROCESS.
  (make <process>
    :atoms (apply hash-table 'string=?
                  (map (^(k v) (cons k (list-copy v)))
                       (hash-table-keys (process 'atoms))
                       (hash-table-values (process 'atoms))))
    :proxy (atom-copy (process 'proxy))))

(define (map-atoms process fn :optional functor)
  ;; Apply FN for all atoms in PROCESS to make a list of results. When
  ;; optional arg FUNCTOR is specified, apply FN for all atoms whose
  ;; functor is FUNCTOR.
  (if (undefined? functor)
      (apply append! (hash-table-map (process 'atoms) (^(k v) (map fn v))))
      (map fn (process functor))))

(define (head-atom proc :optional (atom (process-get proc)) (known-atoms (make-process)))
  ;; Search for an atom considered as a root of a tree in the PROCESS,
  ;; which ATOM is a member of the tree. When ATOM is omitted or
  ;; #<undef>, search for an arbitrary head-atom.
  (if (= (atom 'arity) 0) ;; no parent
      atom
      (let1 parent (let1 p (atom (- (atom 'arity) 1))
                     (unless (undefined? p) (p 'atom)))
        (cond [(or (undefined? parent)
                   (not (process-member proc parent))) ;; connected to somewhere outside
               atom]
              [(process-member known-atoms atom) parent] ;; cycle
              [else (process-push! known-atoms atom)
                    (head-atom proc parent known-atoms)]))))

;;   + process <-> sexp

(define (process->sexp process)
  ;;  Generate a (human-readable) sexp from PROCESS. PROCESS is
  ;; denoted with trees as well as possible. freelinks are denoted
  ;; with integers, and other internal links are denoted with unique
  ;; non-integer symbols. When some free ports of the process is
  ;; directly connected, they are added as a list.
  (let ([tmp-proc (process-copy process)] ;; we may destruct this process
        [links ()] ;; register for pending links
        [res ()])  ;; register for resulting sexps
    (define (tree->sexp head-atom :optional (headp #t))
      ;;  Extract a tree whose head is HEAD-ATOM.
      (process-pop! tmp-proc head-atom)
      (cons (head-atom 'name)
            (map-ports head-atom
                       (^p (let1 arg (partner p)
                             (cond [(or (undefined? arg)
                                        (not (process-member process (arg 'atom)))) ;; freelink
                                    (rlet1 portnum (let/cc return
                                                     (dotimes (n (tmp-proc 'arity))
                                                       (when (equal? (tmp-proc (list n)) p)
                                                         (return n))))
                                      (unless (integer? portnum)
                                        (lmn-error "non-linear links in the process."))
                                      (set! (tmp-proc (list portnum)) #f))]
                                   [(assoc-ref links p #f) => ;; pending internal link
                                    (^s (set! links (alist-delete! p links)) s)]
                                   [(or (not (= (arg 'nth) (- ((arg 'atom) 'arity) 1))) ;; not a child
                                        (not (process-member tmp-proc (arg 'atom)))) ;; cycle found
                                    (rlet1 symbol (gensym "L")
                                      (push! links (cons arg symbol)))]
                                   [else ;; proper child
                                    (tree->sexp (arg 'atom) #f)])))
                       0 (- (head-atom 'arity) (if headp 0 1)))))
    (define (links->sexp)
      ;;  Extract remaining links in process
      (rlet1 res ()
        (dotimes (n (tmp-proc 'arity))
          (if-let1 arg (tmp-proc (list n))
            (begin (when (undefined? arg)
                     (error "links->sexp: unassigned port in the process."))
                   (unless (eq? (arg 'atom) ((partner arg) 'atom))
                     (error "links->sexp: port outside the process is assigned."))
                   (push! res (list n (arg 'nth)))
                   (set! (tmp-proc (list n)) #f)
                   (set! (tmp-proc (list (arg 'nth))) #f))))))
    (while (process-get tmp-proc) => atom
           (push! res (tree->sexp (head-atom process atom))))
    (append! res (links->sexp))))

(define (sexp->process trees)
  ;;  Construct process from a sexp.
  (let* ([links ()] ;; register for pending links
         [res (make-process (+ 1 ((rec (max* list)
                                    (cond [(integer? list) list]
                                          [(pair? list) (apply max (map max* list))]
                                          [else -1])) trees)))])
    (define (tree->process tree :optional (parent-port #f))
      (if (pair? tree)
          (cond [(string? (car tree)) ;; atom
                 (let* ([args (cdr tree)]
                        [children (length args)]
                        [arity (+ children (if parent-port 1 0))]
                        [atom (make-atom (car tree) arity)])
                   (when parent-port
                     (connect atom (- arity 1)
                              (parent-port 'atom) (parent-port 'nth)))
                   (process-push! res atom)
                   (dotimes (n children)
                     (tree->process (car args) (atom (list n)))
                     (set! args (cdr args))))]
                [(integer? (car tree)) ;; direct link
                 (unless (and (integer? (cadr tree))
                              (= (length tree) 2))
                   (error "tree->process: direct link must be list of two integers."))
                 (connect res (list (car tree)) res (list (cadr tree)))]
                [else
                 (error #`"tree->process: atom name must be string (,(car tree))")])
          (cond [(integer? tree) ;; port
                 (unless parent-port
                   (error "port must be connected to somewhere (try: 1 -> (\"1\"))."))
                 (set! (res (list tree)) parent-port)]
                [(symbol? tree) ;; internal link
                 (unless parent-port
                   (error "link must be connected to somewhere (try: a -> (a))."))
                 (if-let1 link (assq-ref links tree)
                   (begin (set! links (alist-delete! tree links eq?))
                          (connect (link 'atom) (link 'nth)
                                   (parent-port 'atom) (parent-port 'nth)))
                   (push! links (cons tree parent-port)))]
                [else
                 (error "tree->process: link must be a symbol (try: \"a\" -> (\"a\")).")])))
    (map tree->process trees)
    res))

(define (process-deep-copy process)
  ;;  Copy all atoms in PROCESS and return a new process.
  (sexp->process (process->sexp process)))

;;   + atom/process common operations

;; functions defined here are also for processes

(define (map-args atom-or-proc fn :optional (from 0) (to (atom-or-proc 'arity)))
  ;; Call FN for all arguments of ATOM-OR-PROC to make a list of
  ;; results. When optional arg FROM and TO are specified, FN is
  ;; applied only from FROMth argument to TOth argument.
  (map (^n (fn (atom-or-proc n))) (iota to from)))

(define (map-ports atom-or-proc fn :optional (from 0) (to (atom-or-proc 'arity)))
  ;; Call FN for all ports of ATOM-OR-PROC to make a list of
  ;; results. Ports are passed to FN as <atomport> objects.
  (map (^n (fn (atom-or-proc (list n)))) (iota to from)))

(define (connect atom-or-proc1 n1 atom-or-proc2 n2)
  ;;  Connect (ATOM-OR-PROC1 N1) and (ATOM-OR-PROC2 N2).
  (set! (atom-or-proc1 n1) (atom-or-proc2 (list n2)))
  (set! (atom-or-proc2 n2) (atom-or-proc1 (list n1))))

;; **FIXME** DIRECT LINKS AND "CONNECT"
;;
;; "connect" works correctly ONLY ONCE for a direct link in a process
;;
;; ex. a(A), b(B)
;;
;;     +-$p-----+         +-$p-----+               +-$p-----+
;;     |   0->a |         |   0->a |               |   0->a |
;;     |b       |         |b     ^ |               |b<-+  ^ |
;;     |^       |  A=foo  |^     | |        B=bar  |^  |  | |
;;     ||       x   ==>   ||     +---->foo   ==>   ||  |  +---->foo
;;     |1       |A        |1       |A              |1  ++   |A
;;     +---x----+         +---x----+               +----|---+
;;          B                  B                        |B
;;                                                      v
;;     * 0, 1 are the ports of the proxy atom          bar
;;
;; ex. $p[foo, bar] where $p[A, B] = { A = B }
;;
;;     +-$p----+         +-$p----+               +-$p----+
;;     | +--->0|         | +----0|               | +----0|
;;     | |     |  A=foo  | |     |        B=bar  | |     |
;;     | |     x   ==>   | |   +---->foo   ==>   | |   +---->foo
;;     | v     |A        | v   | |A              | v   | |A   ^
;;     | 1     |         | 1<--+ |               | 1---+ |    |
;;     +---x---+         +---x---+               +---x---+    |
;;          B                 B                       B bar<--+
;;

(define (connected? atom-or-proc1 n1 atom-or-proc2 n2)
  ;;  Check if (ATOM-OR-PROC1 N1) and (ATOM-OR-PROC2 N2) are connected.
  (let ([port1 (atom-or-proc1 n1)]
        [port2 (atom-or-proc2 n2)])
    (and (equal? (partner port1) port2)
         (equal? (partner port2) port1))))

(define-test "connect and atom-copy with internal links"
  (let ([atom1 (make-atom "foo" 3)]
        [atom2 (make-atom "bar" 1)])
    (connect atom1 0 atom2 0)
    (connect atom1 1 atom1 2)
    (connected? atom1 1 atom1 2) :: true?
    (let1 atom1? (atom-copy atom1)
      (atom1? 'functor) ==> (atom1 'functor)
      (connected? atom1 1 atom1 2) :: true?
      (connected? atom1? 1 atom1? 2) :: true?
      (equal? (atom1? 0) (atom2 '(0))) :: true?)))

;;   + tests / examples

(define-test "basic methods of processes"
  (let ([process (make-process 2)]
        [atom1 (make-atom "x" 1)]
        [atom2 (make-atom "foo" 2)]
        [atom3 (make-atom "bar" 2)]
        [atom4 (make-atom "y" 1)])
    ;; x -- foo -- bar -- y
    (connect atom1 0 atom2 0)
    (connect atom2 1 atom3 0)
    (connect atom3 1 atom4 0)
    (connected? atom1 0 atom2 0) :: true?
    (connected? atom2 0 atom1 0) :: true?
    (connected? atom1 0 atom2 1) ==> #f
    (map-args atom2 (^p ((p 'atom) 'name))) ==> '("x" "bar")
    ;; { x -- foo -- bar -- y }
    (process-push! process atom1 atom2 atom3 atom4)
    (process-member process atom1) :: true?
    (process-member process (make-atom "x" 1)) ==> #f
    ;; x -{- foo -- bar -}- y
    (process-pop! process "x_1") ==> atom1
    (set! (process '(0)) (atom2 '(0)))
    (process-pop! process atom4) ==> atom4
    (set! (process '(1)) (atom3 '(1)))
    (process-member process atom1) ==> #f
    (process-member process atom2) :: true?
    (process-member process atom3) :: true?
    (process-member process atom4) ==> #f
    (map-atoms process (^a (a 'arity))) ==> '(2 2)
    (map-args process (^p ((p 'atom) 'name))) ==> '("x" "y")
    ;; a -{- foo -- bar -}- b
    (let ([atoma (make-atom "a" 1)]
          [atomb (make-atom "b" 1)])
      (connect process 0 atoma 0)
      (connect process 1 atomb 0)
      (connected? atom2 0 atoma 0) :: true?
      (connected? process 1 atomb 0) :: true?)))

(define-test "process-copy / process-pop!"
  (let ([proc (make-process 1)]
        [atom1 (make-atom "foo" 0)]
        [atom2 (make-atom "bar" 1)]
        [atom3 (make-atom "baz" 1)])
    (process-push! proc atom1 atom2)
    (set! (proc '(0)) (atom2 '(0)))
    (process-get proc "foo_0") ==> atom1
    (let1 proc2 (process-copy proc)
      (process-member proc2 atom1) :: true?
      (process-member proc2 atom2) :: true?
      (process-member proc2 atom3) ==> #f
      (process-pop! proc2 atom2) ==> atom2
      (process-push! proc2 atom3)
      (set! (proc2 '(0)) (atom3 '(0)))
      (process-member proc atom3) ==> #f
      (process-member proc atom2) :: true?
      (proc '(0)) ==> (atom2 '(0))
      (proc2 '(0)) ==> (atom3 '(0))
      (process-pop! proc2 atom3) :: true?
      (process-pop! proc2) ==> atom1
      (process-pop! proc2) ==> #f)))

(define-test "freelinks connected to nowhere"
  (let ([proc (make-process 3)]
        [atom1 (make-atom "a" 1)]
        [atom2 (make-atom "x" 1)]
        [atom3 (make-atom "y" 1)]
        [atom4 (make-atom "z" 1)])
    ;; +-------+
    ;; | a  +--+--2
    ;; | |  |  |
    ;; +-+--+--+
    ;;   |  |
    ;;   0  1
    (set! (proc '(0)) (atom1 '(0)))
    (connect proc '(1) proc '(2))
    ;; +-------+
    ;; | a  +--+--z
    ;; | |  |  |
    ;; +-+--+--+
    ;;   |  |
    ;;   x  y
    (connect proc 0 atom2 0)
    (connect proc 1 atom3 0)
    (connect proc 2 atom4 0)
    (connected? atom1 0 atom2 0) :: true?
    (connected? atom3 0 atom4 0) :: true?))

(define-test "head-atom"
  (let ([process (make-process 0)]
        [atom1 (make-atom "foo" 4)]
        [atom2 (make-atom "foo" 2)]
        [atom3 (make-atom "bar" 1)]
        [atom4 (make-atom "baz" 1)])
    ;; foo(bar, baz, foo(L1), L1)
    ;; foo(foo(bar, baz, L1), L1)
    (connect atom1 0 atom3 0)
    (connect atom1 1 atom4 0)
    (connect atom1 2 atom2 1)
    (connect atom1 3 atom2 0)
    (process-push! process atom1)
    (process-push! process atom2)
    (process-push! process atom3)
    (process-push! process atom4)
    ((head-atom process atom4) 'name) ==> "foo"))

;;   + tests / examples

(define-test "sexp->process a closed process"
  (let* ([proc (sexp->process
                '(("a" L1 ("b") L2 ("c" L2)) ("d" L3 L3) ("e" L1) ("f")))]
         [a (process-get proc "a_4") :: true?]
         [b (process-get proc "b_1") :: true?]
         [c (process-get proc "c_2") :: true?]
         [d (process-get proc "d_2") :: true?]
         [e (process-get proc "e_1") :: true?]
         [f (process-get proc "f_0") :: true?])
    (connected? a 0 e 0) :: true?
    (connected? a 1 b 0) :: true?
    (connected? a 2 c 0) :: true?
    (connected? a 3 c 1) :: true?
    (connected? d 0 d 1) :: true?
    (proc 'arity) ==> 0))

(define-test "sexp->process a process with freelinks"
  (let* ([proc (sexp->process
                '(("a" ("b" 0) L1 ("c" L1 1)) (2 3)))]
         [a (process-get proc "a_3")]
         [b (process-get proc "b_2")]
         [c (process-get proc "c_3")])
    (connected? a 0 b 1) :: true?
    (connected? a 1 c 0) :: true?
    (connected? a 2 c 2) :: true?
    (proc 0) :: undefined?
    (proc '(0)) ==> (b '(0))
    (proc '(1)) ==> (c '(1))
    (proc 2) ==> (proc '(3))))

(define-test "process->sexp a closed process"
  (let* ([proc (sexp->process
                (process->sexp
                 (sexp->process
                  '(("a" ("e") ("b") L1 L2) ("c" L1 L2) ("d" L2 L2) ("f")))))]
         [a (process-pop! proc "a_4") :: true?]
         [b (process-pop! proc "b_1") :: true?]
         [c (process-pop! proc "c_2") :: true?]
         [d (process-pop! proc "d_2") :: true?]
         [e (process-pop! proc "e_1") :: true?]
         [f (process-pop! proc "f_0") :: true?])
    (process-pop! proc) ==> #f
    (connected? a 0 e 0) :: true?
    (connected? a 1 b 0) :: true?
    (connected? a 2 c 0) :: true?
    (connected? a 3 c 1) :: true?
    (connected? d 0 d 1) :: true?))

(define-test "process->sexp a process with freelinks"
  (let* ([proc (sexp->process
                (process->sexp
                 (sexp->process
                  '(("a" 1 ("b") ("c" ("d")) 3) (0 2)))))]
         [a (process-pop! proc "a_4") :: true?]
         [b (process-pop! proc "b_1") :: true?]
         [c (process-pop! proc "c_2") :: true?]
         [d (process-pop! proc "d_1") :: true?])
    (process-pop! proc) ==> #f
    (connected? proc 0 proc 2) :: true?
    (proc '(1)) ==> (a '(0))
    (proc '(3)) ==> (a '(3))))

(define-test "process-deep-copy"
  (let* ([proc1 (sexp->process '(("a" ("b") 0) ("c" 1 L1) ("d" L1 2)))]
         [proc2 (process-deep-copy proc1)]
         [a (process-pop! proc1 "a_2") :: true?]
         [b (process-pop! proc1 "b_1") :: true?]
         [c (process-pop! proc1 "c_2") :: true?]
         [d (process-pop! proc1 "d_2") :: true?]
         [a2 (process-pop! proc2 "a_2") :: true?]
         [b2 (process-pop! proc2 "b_1") :: true?]
         [c2 (process-pop! proc2 "c_2") :: true?]
         [d2 (process-pop! proc2 "d_2") :: true?])
    (proc1 '(0)) ==> (a '(1))
    (proc1 '(1)) ==> (c '(0))
    (proc1 '(2)) ==> (d '(1))
    (proc2 '(0)) ==> (a2 '(1))
    (proc2 '(1)) ==> (c2 '(0))
    (proc2 '(2)) ==> (d2 '(1))
    (equal? a a2) ==> #f
    (equal? b b2) ==> #f
    (equal? c c2) ==> #f
    (equal? d d2) ==> #f))

;; + rules, types
;;   + rules, types

;; example:
;;
;; foo(L1, bar($x), baz(L1, L2)), qux(L3, L4), $y(L2, L3) :- {
;;        rel($x, $y) | foo($x), bar(L5), $y(L5, L4).
;;     p1($x), p2($y) | foo(L4).
;; }

;; compiled rule:
;;
;; rule {
;;   spec: (7 . 6) <-- minimum register size
;;                 l0        l1         l2 l3          l4  l5
;;   trees: ("foo" #f ("bar" #f) ("baz" #f #f)) ("qux" #f, #f)
;;          |----------------p1---------------| |-----p2-----|
;;
;;   links: (0 2) <-- link 0 and 2 are connected
;;
;;   clauses:
;;       {
;;           contexts: (<rel> 1 3 4) <-- l1, l3, l4 must be connected to <rel> process
;;                     |----p3-----|
;;
;;           split: (3 (1) (3 4))
;;                    |p4| |-p5|
;;                                    l6
;;           body: ("foo" (2)) ("bar" #f) (3 6 5)
;;       }
;;       {
;;           contexts: (<p1> 1) (<p2> 3 4)
;;                     |--p3--| |---p4---|
;;
;;           body: ("foo" (5))
;;       }
;; }

(define-class <rule> ()
  ((spec :init-keyword :spec)
   (trees :init-keyword :trees :init-value ())
   (links :init-keyword :links :init-value ())
   (clauses :init-keyword :clauses)))

(define-class <rule-clause> ()
  ((contexts :init-keyword :contexts :init-value ())
   (split :init-keyword :split :init-value ())
   (body :init-keyword :body :init-value ())))

(define-method object-apply ((rule <rule>) (ref <symbol>))
  (slot-ref rule ref))

(define-method object-apply ((clause <rule-clause>) (ref <symbol>))
  (slot-ref clause ref))

;; type int_list(X) { '[]'(X). '.'(H, T, X) :- int(H), int_list(T). }.
;;              l0
;; {
;;     clauses:
;;         {
;;             spec: (2 . 1)
;;                          l1
;;             trees: ("[]" #f)
;;                    |---p0--|
;;             links: (0 . 1)
;;         ;;  head-atom: (0)  <- head of the tree must be connected to l0
;;         }
;;         {
;;             spec: (4 . 3)
;;                          l1  l2  l3
;;             trees: (".", #f, #f, #f)
;;                    |-------p0------|
;;             links: (0 . 3)
;;             goals: ("int" 1) ("int_list" 2)
;;                    |------p1------| |---------p2--------|
;;         ;;  head-atom: (0)
;;         }
;; }

;; type foo(X, Y) { foo(a(X) , b(Y), Z) :- bar(Z). }
;;         l0 l1
;; {
;;     clauses:
;;         {
;;             spec: (9 . 4)
;;                         l2 l3       l4 l5         l6 l7 l8
;;             trees: ("a" #f #f) ("b" #f #f) ("foo" #f #f #f)
;;                    |----p0---| |----p1---| |------p2------|
;;             links: (0 . 2) (1 . 4) (3 . 6) (5 . 7)
;;             goals: ("bar" 8)
;;                    |------p3------|
;;             head-atom: (0 1 #f)
;;         }
;; }

(define-class <type> ()
  ((clauses :init-keyword :clauses)))

;; **FIXME** IMPLEMENT HEAD_ATOM OPTIMIZATION
(define-class <type-clause> ()
  ((spec :init-keyword :spec)
   (trees :init-keyword :trees :init-value ())
   (links :init-keyword :links :init-value ())
   (goals :init-keyword :goals :init-value ())
   ;; (head-atom :init-keyword :head-atom :init-value ())
   ))

(define-method object-apply ((rule <type>) (ref <symbol>))
  (slot-ref rule ref))

(define-method object-apply ((clause <type-clause>) (ref <symbol>))
  (slot-ref clause ref))

(define (rule->string rule)
  (if (procedure? rule)
      (x->string rule)
      (string-append/shared
       "compiled-rule ::\n"
       "    spec : " (x->string (slot-ref rule 'spec)) "\n"
       "   trees : " (x->string (slot-ref rule 'trees)) "\n"
       "   links : " (x->string (slot-ref rule 'links)) "\n"
       " clauses : \n"
       (string-concatenate
        (map (^c (string-append/shared
                  "     {\n"
                  "       contexts : " (x->string (slot-ref c 'contexts)) "\n"
                  "          split : " (x->string (slot-ref c 'split)) "\n"
                  "           body : " (x->string (slot-ref c 'body)) "\n"
                  "     }\n"))
             (slot-ref rule 'clauses))))))

(define (type->string type)
  (string-append/shared
   "compiled-type ::\n"
   " clauses : \n"
   (string-concatenate
    (map (^c (string-append/shared
              "     {\n"
              "        spec : " (x->string (slot-ref c 'spec)) "\n"
              "       trees : " (x->string (slot-ref c 'trees)) "\n"
              "       links : " (x->string (slot-ref c 'links)) "\n"
              "       goals : " (x->string (slot-ref c 'goals)) "\n"
              "     }\n"))
         (slot-ref type 'clauses)))))

;;   + rule application
;;   + | utilities

(define (>> :rest fns)
  ;;  Compose 4-ary (proc lreg preg type-env :optional next) functions.
  (^(proc lreg preg type-env :optional next)
    (cond [(null? fns) (undefined)]
          [else ((car fns) proc lreg preg type-env
                 (^() ((apply >> (cdr fns)) proc lreg preg type-env)))])))

(define (>or :rest fns)
  ;;  Compose 4-ary (proc lreg preg type-env :optional next) functions.
  (^(proc lreg preg type-env :optional next)
    (let ([lreg-state (register-get-state lreg)]
          [preg-state (register-get-state preg)])
      (let/cc return
        (dolist (fn fns)
          (cond [(and (fn proc lreg preg type-env)
                      (or (undefined? next) (next)))
                 (return #t)]
                [else
                 (register-restore-state! lreg lreg-state)
                 (register-restore-state! preg preg-state)]))
        #f))))

;; **FIXME** "NEXT" CANNOT FAIL
(define (rule-delete-processes!)
  (^(proc lreg preg type-env :optional next)
    ;;  Delete all atoms in PREG. When optional argument NEXT is
    ;; specified, it is called with no arguments at last.
    (dump +1 ">> delete process.")
    (map-register preg
                  (^p (map-atoms p (^a (process-pop! proc a)
                                       (dump 0 "delete " (a 'functor) ".")))))
    (dump -1 "<< delete done.")
    (or (undefined? next) (next))))

;;   + | register

(define-class <register> ()
  ((stack :init-keyword :vector)
   (ptr :init-keyword :ptr)))

(define-method object-apply ((register <register>) (n <integer>))
  (when (< n (slot-ref register 'ptr))
    (vector-ref (slot-ref register 'stack) n)))

(define (make-register n)
  (make <register> :vector (make-vector n) :ptr 0))

(define (register-push! register item)
  (let ([stack (slot-ref register 'stack)]
        [ptr (slot-ref register 'ptr)])
    (when (= ptr (vector-length stack))
      (error "register-push!: register is full."))
    (vector-set! stack ptr item)
    (slot-set! register 'ptr (+ 1 ptr))))

(define (register-pop! register)
  (let ([stack (slot-ref register 'stack)]
        [ptr (slot-ref register 'ptr)])
    (when (= ptr 0)
      (error "register-pop!: register is empty."))
    (slot-set! register 'ptr (- ptr 1))
    (vector-ref stack (- ptr 1))))

(define (register-get-state register)
  ;;  Get REGISTER's state so that we may go back to the state.
  (slot-ref register 'ptr))

(define (register-restore-state! register state)
  ;;  Restore REGISTER's state to STATE.
  (slot-set! register 'ptr state))

(define (map-register register fn)
  ;;  Apply FN for all items in REGISTER.
  (let1 vec (slot-ref register 'stack)
    (map (^n (fn (vector-ref vec n)))
         (iota (slot-ref register 'ptr)))))

(define-test "register"
  (let1 register (make-register 5)
    (register-push! register 'a)
    ;; a
    (let1 state (register-get-state register)
      (register-push! register 'b)
      (register-push! register 'c)
      (register 0) ==> 'a
      (register 1) ==> 'b
      (register 2) ==> 'c
      (map-register register identity) ==> '(a b c)
      (register-restore-state! register state))
    ;; a
    (register 0) ==> 'a
    (register 1) ==> (undefined)
    (register 2) ==> (undefined)
    (map-register register identity) ==> '(a)
    ;; a d e
    (register-push! register 'd)
    (register-push! register 'e)
    (register 0) ==> 'a
    (register 1) ==> 'd
    (register 2) ==> 'e
    (map-register register identity) ==> '(a d e)))

(define (register-process-member register atom)
  ;;  Check if ATOM is contained in REGISTER. All items in REGISTER must be process.
  (let/cc return
    (map-register register
                  (^p (if (process-member p atom) (return #t))))
    #f))

(define-test "register-process-member"
  (let ([reg (make-register 5)]
        [proc1 (make-process)]
        [proc2 (make-process)]
        [proc3 (make-process)]
        [atom1 (make-atom "foo" 0)]
        [atom2 (make-atom "foo" 0)]
        [atom3 (make-atom "foo" 0)])
    (process-push! proc1 atom1)
    (process-push! proc2 atom2)
    (process-push! proc3 atom3)
    (register-push! reg proc1)
    (register-push! reg proc2)
    (register-push! reg proc3)
    (register-process-member reg atom1) ==> #t
    (register-process-member reg atom2) ==> #t
    (register-process-member reg atom3) ==> #t
    (register-process-member reg (make-atom "foo" 0)) ==> #f))

;;   + | trees

(define (rule-match-tree tree :optional head-atom)
  (^(proc lreg preg type-env :optional next)
    ;;  Search for a set of atoms in PROC that matches TREE, and
    ;; return whether succeeded. When HEAD-ATOM is given, it must be
    ;; the root of the tree.
    ;;  All found ports are pushed to LREG, and the set of atoms is
    ;; pushed to PREG as a 0-ary process. If PREG already has some
    ;; members, atoms in it will not match.
    ;;  When optional arg NEXT is specified, it is called after matching
    ;; without arguments and if it return #f search for another tree.
    (define (rule-match-tree-1 tree head-atom
                               :optional (acc (make-process)) (rootp #t))
      ;;  Recursive function for rule-match-tree. Collect atoms in ACC.
      (let/cc return
        (let ([children (- (head-atom 'arity) (if rootp 0 1))]
              [tree-args (cdr tree)])
          (and (not (register-process-member preg head-atom))
               (process-push! acc head-atom)
               (string=? (head-atom 'name) (car tree))
               (= children (length tree-args))
               (dotimes (n children)
                 (cond [(car tree-args) ;; atom
                        (unless (and (= ((head-atom n) 'nth)
                                        (- (((head-atom n) 'atom) 'arity) 1))
                                     (rule-match-tree-1
                                      (car tree-args) ((head-atom n) 'atom) acc #f))
                          (return #f))]
                       [else ;; link
                        (register-push! lreg (head-atom n))])
                 (set! tree-args (cdr tree-args)))
               acc))))
    (let ([lreg-state (register-get-state lreg)]
          [preg-state (register-get-state preg)])
      (dump +1 ">> match " tree ".")
      (let/cc return
        (dolist (candidate (if (undefined? head-atom)
                               (proc #`",(car tree)_,(length (cdr tree))")
                               (list head-atom)))
          (let1 res (rule-match-tree-1 tree candidate)
            (cond [(and res
                        (begin (register-push! preg res)
                               (dump 0 "a match found.")
                               (or (undefined? next) (next))))
                   (dump -1 "<< match succeeded")
                   (return #t)]
                  [else
                   (register-restore-state! lreg lreg-state)
                   (register-restore-state! preg preg-state)])))
        (dump -1 "XX no matches.")
        #f))))

(define-test "rule-match-tree (1)"

  (let ([proc (sexp->process '(("foo" ("a" ("c") ("d")))))]
        [lreg (make-register 2)]
        [preg (make-register 1)])
    ((rule-match-tree '("foo" ("b" #f #f))) proc lreg preg #f) ==> #f)

  (let ([proc (sexp->process '(("foo" ("b" ("b") ("c")))))]
        [lreg (make-register 2)]
        [preg (make-register 1)])
    ((rule-match-tree '("foo" ("b" #f #f))) proc lreg preg #f) :: true?
    (((lreg 0) 'atom) 'name) ==> "b"
    (((lreg 1) 'atom) 'name) ==> "c"
    (process-pop! (preg 0) "foo_1") :: true?
    (process-pop! (preg 0) "b_3") :: true?
    (process-pop! (preg 0)) ==> #f)
  )

;;   + | links

(define (rule-check-links links)
  (^(proc lreg preg type-env :optional next)
    ;;  Check all LINKS are correctly connected wrt LREG.
    ;;  If NEXT is specified, return value for success is
    ;; (next). Otherwise a boolean value.
    (dump +1 ">> check links " links ".")
    (let/cc return
      (dolist (link links)
        (let ([port1 (lreg (car link))]
              [port2 (lreg (cdr link))])
          (unless (connected? (port1 'atom) (port1 'nth)
                              (port2 'atom) (port2 'nth))
            (dump -1 "XX check links failed.")
            (return #f))))
      (dump -1 "<< check links succeeded.")
      (or (undefined? next) (next)))))

(define-test "rule-match-tree, check-links"
  (let1 tree '("a" ("b" #f ("d")) #f)

    ;; match (not connected)
    (let* ([proc (sexp->process '(("a" ("b" ("c") ("d")) ("e"))))]
           [lreg (make-register 2)]
           [preg (make-register 1)])
      ((rule-match-tree tree) proc lreg preg #f) :: true?
      ((rule-check-links '((0 . 1))) proc lreg preg #f) ==> #f
      (let1 p (preg 0)
        (process-pop! p "a_2") :: true?
        (process-pop! p "b_3") :: true?
        (process-pop! p "d_1") :: true?
        (process-pop! p) ==> #f)
      (map-register lreg (^p ((p 'atom) 'name))) ==> '("c" "e"))

    ;; match (connected)
    (let ([proc (sexp->process '(("a" ("b" L1 ("d")) L1)))]
          [lreg (make-register 2)]
          [preg (make-register 1)])
      ((rule-match-tree tree) proc lreg preg #f) :: true?
      ((rule-check-links '((0 . 1))) proc lreg preg #f) :: true?)

    ;; mismatch (argument order)
    (let1 proc (sexp->process '(("a" L1 ("e")) ("b" L1 ("c") ("d"))))
      ((rule-match-tree tree) proc (make-register 2) (make-register 1) #f)
      ==> #f)

    ;; mismatch (argument number)
    (let1 proc (sexp->process '(("a" ("b" ("c") ("d") ("f")) ("e"))))
      ((rule-match-tree tree) proc (make-register 2) (make-register 1) #f)
      ==> #f)

    ;; backtracking (match)
    (let1 proc (sexp->process '(("a" ("b" L2 ("d")) L3)
                                ("a" ("b" L1 ("d")) L1)
                                ("a" ("b" L2 ("d")) L3)))
      ((>> (rule-match-tree tree)
           (rule-match-tree tree)
           (rule-match-tree tree)
           (rule-check-links '((0 . 1) (2 . 4) (3 . 5))))
       proc (make-register 6) (make-register 3) #f) :: true?)

    ;; backtracking (mismatch)
    (let1 proc (sexp->process '(("a" ("b" L2 ("d")) L3)
                                ("a" ("b" ("c") ("d")) ("e"))
                                ("a" ("b" L2 ("d")) L3)))
      ((>> (rule-match-tree tree)
           (rule-match-tree tree)
           (rule-match-tree tree)
           (rule-check-links '((0 . 1) (2 . 4) (3 . 5))))
       proc (make-register 6) (make-register 3) #f) ==> #f)
    ))

;;   + | built-in types

(define (rule-traverse-contexts types)
  (^(proc lreg preg type-env :optional next)
    ;;  Check if the TYPES are satisfied, and push the processes to
    ;; PREG. If PREG has already some members, atoms in it will not
    ;; match.
    ;;  If NEXT is specified, return value for success is
    ;; (next). Otherwise a boolean value.
    (let1 preg-state (register-get-state preg)
      (dump +1 #`">> type check : ,|types|.")
      (let/cc return
        (dolist (type types)
          (dump 0 "check " type ".")
          (if-let1 res (type-traverse-context
                        proc preg (car type) type-env (map lreg (cdr type)))
            (begin (register-push! preg res)
                   (dump 0 "succeeded " (process->sexp res) " : " (car type) "."))
            (begin (register-restore-state! preg preg-state)
                   (dump -1 #`"XX type check failed : ,|type|.")
                   (return #f))))
        (dump -1 "<< type check succeeded.")
        (rlet1 ret (or (undefined? next) (next))
          (unless ret (register-restore-state! preg preg-state)))))))

(define (type-traverse-context proc preg typename type-env ports)
  ;;  Search for a subgraph whose ports are exactly PORTS, and
  ;; satisfies TYPE. If any, return it as a process. Otherwise return
  ;; #f. All atoms in the subgraph must be of PROC, and not in PREG.
  (if-let1 type (hash-table-get type-env typename #f)
    (cond [(procedure? type) ;; built-in types : (Proc, Preg, Port, ...) -> Maybe Proc
           (apply type proc preg ports)]
          [(is-a? type <type>)
           (apply type-user-defined proc preg type type-env ports)]
          [else
           (error "type-traverse-context: invalid type constraint.")])
    (lmn-error #`"unknown type ,|typename|.")))

(define (type-subr-linked proc preg port1 port2)
  ;;  Check if PORT1 and PORT2 are connected.
  (and (connected? (port1 'atom) (port1 'nth)
                   (port2 'atom) (port2 'nth))
       (rlet1 newproc (make-process 2)
         (connect newproc '(0) newproc '(1)))))

(define (type-subr-int proc preg port)
  ;;  Check if PORT is connected to a int atom in PROC. If so,
  ;; return an unary process whose member is the atom. Otherwise
  ;; return #f.  Atoms in PREG will not match.
  (let1 atom (port 'atom)
    (and (not (register-process-member preg atom))
         (process-member proc atom)
         (= (atom 'arity) 1)
         (integer? (string->number (atom 'name)))
         (rlet1 newproc (make-process 1)
           (process-push! newproc atom)
           (set! (newproc '(0)) (atom '(0)))))))

(define (type-subr-ints-compare fn proc preg port1 port2)
  ;;  Check if PORT1 and PORT2 are connected to int atoms in PROC,
  ;; and the values satisfy (FN val1 val2). If so, return a 2-ary
  ;; process of the two int atoms. Otherwise return #f. Atoms in
  ;; PREG will not match.
  (let ([atom1 (port1 'atom)]
        [atom2 (port2 'atom)])
    (and (not (eq? atom1 atom2))
         (not (register-process-member preg atom1))
         (not (register-process-member preg atom2))
         (process-member proc atom1)
         (process-member proc atom2)
         (= (atom1 'arity) (atom2 'arity) 1)
         (and-let* ([int1 (string->number (atom1 'name))]
                    [int2 (string->number (atom2 'name))])
           (and (integer? int1) (integer? int2)
                (fn int1 int2)
                (rlet1 newproc (make-process 2)
                  (process-push! newproc atom1)
                  (process-push! newproc atom2)
                  (set! (newproc '(0)) (atom1 '(0)))
                  (set! (newproc '(1)) (atom2 '(0)))))))))
(define type-subr-le (pa$ type-subr-ints-compare <=))
(define type-subr-lt (pa$ type-subr-ints-compare <))
(define type-subr-ge (pa$ type-subr-ints-compare >=))
(define type-subr-gt (pa$ type-subr-ints-compare >))
(define type-subr-eq (pa$ type-subr-ints-compare =))

(define (type-subr-atom proc preg :rest ports)
  ;;  Check if PORTS are connected to an atom in PROC whose args are
  ;; exactly PORTS. If so return a (length PORTS)-ary process whose
  ;; member is the atom. Otherwise return #f. Atoms in PREG will not
  ;; match.
  (unless (pair? ports)
    ;; atom process with no arguments must be a compile-error
    (error "wrong number of arguments for type-subr-atom."))
  (let1 atom ((car ports) 'atom)
    (and (not (register-process-member preg atom))
         (process-member proc atom)
         (= (length ports) (atom 'arity))
         (or (= (atom 'arity) 1)
             (apply eq? (map (^p (p 'atom)) ports)))
         (rlet1 newproc (make-process (atom 'arity))
           (process-push! newproc atom)
           (dotimes (n (atom 'arity))
             (set! (newproc (list n)) (atom (list n))))))))

(define (type-subr-eqv proc preg :rest ports)
  ;;  (type-subr-eqv! proc p01 p02 ... p0n p11 p12 ... p1n) checks
  ;; (atom p01 p02 ... p0n) and (atom p11 p12 ... p1n) are different
  ;; but equivalent atoms in PROC. If so, return 2n-ary process of the
  ;; two atmos. Otherwise return #f. Atoms in PREG will not match.
  (let* ([arity (length ports)]
         [harity (div arity 2)])
    (unless (= arity (+ harity harity))
      (lmn-error "wrong number of arguments for =."))
    (receive (a b) (split-at! ports harity)
      (let* ([atom1 ((car a) 'atom)]
             [atom2 ((car b) 'atom)])
        (and (not (eq? atom1 atom2))
             (not (register-process-member preg atom1))
             (not (register-process-member preg atom2))
             (process-member proc atom1)
             (process-member proc atom2)
             (= harity (atom1 'arity) (atom2 'arity))
             (or (= harity 1)
                 (and (apply eq? (map (^p (p 'atom)) a))
                      (apply eq? (map (^p (p 'atom)) b))))
             (string=? (atom1 'functor) (atom2 'functor))
             (rlet1 newproc (make-process arity)
               (process-push! newproc atom1)
               (process-push! newproc atom2)
               (dotimes (n harity)
                 (set! (newproc (list n)) (atom1 (list n)))
                 (set! (newproc (list (+ n harity))) (atom2 (list n))))))))))

(define (type-subr-ground-internal proc preg ports :optional (connected #f))
  ;;  Check that ports are not connected to either atoms outside PROC
  ;; or atoms in PREG. If so return a (length PORTS)-ary process of
  ;; connected atoms. Otherwise return #f.
  ;;  When optional arg CONNECTED is specified, all ports must
  ;; be connected each other.
  (unless (pair? ports)
    ;; atom process with no arguments must be a compile-error
    (error "wrong number of arguments for type-subr-ground."))
  (let/cc return
    (let* ([args (map partner ports)] ;; args of the ground
           [arity (length ports)]
           [newproc (make-process arity)])
      (define (traverse-port port)
        (let1 atom (port 'atom)
          (cond [(member port args) ;; reached port
                 (set! args (delete1! port args))]
                [(not (process-member newproc atom)) ;; unknown atom found
                 (when (or (not (process-member proc atom))
                           (register-process-member preg atom))
                   (return #f))
                 (process-push! newproc atom)
                 (map-args atom traverse-port)])))
      (let set-ports ([port-nth-pairs (map cons ports (iota arity))])
        (when (pair? port-nth-pairs)
          (let ([port (caar port-nth-pairs)]
                [n (cdar port-nth-pairs)])
            (cond [(assoc (partner port) port-nth-pairs)
                   => (^p (alist-delete! (car p) port-nth-pairs)
                          (connect newproc (list n) newproc (list (cdr p))))]
                  [else
                   (set! (newproc (list n)) port)]))
          (set-ports (cdr port-nth-pairs))))
      (traverse-port (car ports))
      (cond [(null? args) newproc]
            [connected #f]
            [else (map traverse-port (cdr ports)) newproc]))))
(define (type-subr-ground proc preg :rest ports)
  ;;  Check that ports are not connected to either atoms outside PROC
  ;; or atoms in PREG. If so, return a (length PORTS)-ary process of
  ;; connected atoms. Otherwise return #f.
  (type-subr-ground-internal proc preg ports))
(define (type-subr-connected proc preg :rest ports)
  ;;  Check that ports are not connected to either atoms outside PROC
  ;; or atoms in PREG, and all ports are connected (directly or
  ;; not). If so, return a (length PORTS)-ary process of connected
  ;; atoms. Otherwise return #f.
  (type-subr-ground-internal proc preg ports #t))

;; **FIXME** DEPENDS ON hash-table IMPLEMENTATION
(define (type-subr-equal proc preg :rest ports)
  ;;  (type-subr-equal! proc preg p01 p02 ... p0n p11 p12 ... p1n)
  ;; checks (ground p01 p02 ... p0n) and (ground p11 p12 ... p1n) are
  ;; different but equivalent grounds. If so, return 2n-ary process of
  ;; the two grounds. Otherwise return #f.
  (let* ([arity (length ports)]
         [harity (div arity 2)])
    (unless (= arity (+ harity harity))
      (lmn-error "wrong number of arguments for ===."))
    (receive (a b) (split-at! ports harity)
      (let ([p1 (type-subr-ground-internal proc preg a)]
            [p2 (type-subr-ground-internal proc preg b)])
        (and p1 p2
             (equal? (process->sexp p1) (process->sexp p2))
             (rlet1 newproc (make-process arity)
               (map-atoms p1 (pa$ process-push! newproc))
               (map-atoms p2 (pa$ process-push! newproc))
               (dotimes (n harity)
                 (set! (newproc (list n)) (p1 (list n)))
                 (set! (newproc (list (+ n harity))) (p2 (list n))))))))))

(define (default-env)
  (hash-table 'string=?
              `("linked" . ,type-subr-linked)
              `("int" . ,type-subr-int)
              `("<" . ,type-subr-lt)
              `("<=" . ,type-subr-le)
              `(">" . ,type-subr-gt)
              `(">=" . ,type-subr-ge)
              `("=" . ,type-subr-eq)
              `("atom" . ,type-subr-atom)
              `("==" . ,type-subr-eqv)
              `("ground" . ,type-subr-ground)
              `("connected" . ,type-subr-connected)
              `("===" . ,type-subr-equal)))

(define-test "built-in types (atom)"

  (let ([proc (sexp->process '(("a" ("foo") ("bar"))))]
        [lreg (make-register 2)]
        [preg (make-register 7)])
    ((rule-match-tree '("a" #f #f)) proc lreg preg #f)
    ((rule-traverse-contexts '(("linked" 0 1))) proc lreg preg (default-env))
    ==> #f
    ((rule-traverse-contexts '(("int" 0)
                               ("int" 1))) proc lreg preg (default-env))
    ==> #f
    ((rule-traverse-contexts '(("==" 0 1))) proc lreg preg (default-env))
    ==> #f
    ((rule-traverse-contexts '(("atom" 0)
                               ("atom" 1))) proc lreg preg (default-env))
    :: true?
    (process->sexp (preg 1)) ==> '(("foo" 0))
    (process->sexp (preg 2)) ==> '(("bar" 0)))

  (let ([proc (sexp->process '(("a" L1 L1)))]
        [lreg (make-register 2)]
        [preg (make-register 7)])
    ((rule-match-tree '("a" #f #f)) proc lreg preg #f)
    ((rule-traverse-contexts '(("int" 0)
                               ("int" 1))) proc lreg preg (default-env))
    ==> #f
    ((rule-traverse-contexts '(("atom" 0)
                               ("atom" 1))) proc lreg preg (default-env))
    ==> #f
    ((rule-traverse-contexts '(("==" 0 1))) proc lreg preg (default-env))
    ==> #f
    ((rule-traverse-contexts '(("linked" 0 1))) proc lreg preg (default-env))
    :: true?
    (process->sexp (preg 1)) ==> '((0 1)))

  (let ([proc (sexp->process '(("a" ("1") ("2"))))]
        [lreg (make-register 2)]
        [preg (make-register 3)])
    ((rule-match-tree '("a" #f #f)) proc lreg preg #f)
    ((rule-traverse-contexts '(("linked" 0 1))) proc lreg preg (default-env))
    ==> #f
    ((rule-traverse-contexts '(("=" 0 1))) proc lreg preg (default-env))
    ==> #f
    ((rule-traverse-contexts '((">=" 0 1))) proc lreg preg (default-env))
    ==> #f
    ((rule-traverse-contexts '(("int" 0)
                               ("int" 1))) proc lreg preg (default-env))
    :: true?
    (process->sexp (register-pop! preg)) ==> '(("2" 0))
    (process->sexp (register-pop! preg)) ==> '(("1" 0))
    ((rule-traverse-contexts '(("atom" 0)
                               ("atom" 1))) proc lreg preg (default-env))
    :: true?
    (process->sexp (register-pop! preg)) ==> '(("2" 0))
    (process->sexp (register-pop! preg)) ==> '(("1" 0))
    ((rule-traverse-contexts '(("<" 0 1))) proc lreg preg (default-env))
    :: true?
    (process->sexp (preg 1)) ==> '(("1" 0) ("2" 1)))

  (let ([proc (sexp->process '(("a" ("b" ("c")) ("d" L1 L1))))]
        [lreg (make-register 2)]
        [preg (make-register 3)])
    ((rule-match-tree '("a" #f #f)) proc lreg preg #f)
    ((rule-traverse-contexts '(("linked" 0 1))) proc lreg preg (default-env))
    ==> #f
    ((rule-traverse-contexts '(("atom" 0))) proc lreg preg (default-env))
    ==> #f
    ((rule-traverse-contexts '(("atom" 1))) proc lreg preg (default-env))
    ==> #f)

  (let ([proc (sexp->process '(("a" ("b") ("b"))))]
        [lreg (make-register 2)]
        [preg (make-register 7)])
    ((rule-match-tree '("a" #f #f)) proc lreg preg #f)
    ((rule-traverse-contexts '(("linked" 0 1))) proc lreg preg (default-env))
    ==> #f
    ((rule-traverse-contexts '(("int" 0)
                               ("int" 1))) proc lreg preg (default-env))
    ==> #f
    ((rule-traverse-contexts '(("=" 0 1))) proc lreg preg (default-env))
    ==> #f
    ((rule-traverse-contexts '(("atom" 0)
                               ("atom" 1))) proc lreg preg (default-env))
    :: true?
    (process->sexp (register-pop! preg))
    ==> (process->sexp (register-pop! preg)) ==> '(("b" 0))
    ((rule-traverse-contexts '(("==" 0 1))) proc lreg preg (default-env))
    :: true?
    (process->sexp (preg 1)) ==> '(("b" 0) ("b" 1)))
  )

(define-test "built-in types (ground)"

  (let ([proc (sexp->process '(("foo" ("a" ("b")) ("c" ("d")))))]
        [lreg (make-register 2)]
        [preg (make-register 3)])
    ((rule-match-tree '("foo" #f #f)) proc lreg preg #f)
    ((rule-traverse-contexts '(("connected" 0 1))) proc lreg preg (default-env))
    ==> #f
    ((rule-traverse-contexts '(("ground" 0 1))) proc lreg preg (default-env))
    :: true?
    ;; ctx == ("a" ("b") 0) ("c" ("d") 1)
    (let1 ctx (preg 1)
      (let ([a (process-pop! ctx "a_2") :: true?]
            [c (process-pop! ctx "c_2") :: true?])
        (equal? (ctx '(0)) (a '(1))) :: true?
        (equal? (ctx '(1)) (c '(1))) :: true?)
      (process-pop! ctx "b_1") :: true?
      (process-pop! ctx "d_1") :: true?
      (process-pop! ctx) ==> #f))

  (let ([proc (sexp->process '(("foo" ("a" ("b" L1)) ("c" ("d" L1)))))]
        [lreg (make-register 2)]
        [preg (make-register 3)])
    ((rule-match-tree '("foo" #f #f)) proc lreg preg #f)
    (let1 preg-state (register-get-state preg)
      ((rule-traverse-contexts '(("ground" 0 1))) proc lreg preg (default-env))
      :: true?
      (register-restore-state! preg preg-state))
    ((rule-traverse-contexts '(("connected" 0 1))) proc lreg preg (default-env))
    :: true?
    ;; ctx == ("a" ("b" L1) 0) ("c" ("d" L1) 1)
    (let1 ctx (preg 1)
      (let ([a (process-pop! ctx "a_2") :: true?]
            [c (process-pop! ctx "c_2") :: true?])
        (equal? (ctx '(0)) (a '(1))) :: true?
        (equal? (ctx '(1)) (c '(1))) :: true?)
      (let ([b (process-pop! ctx "b_2") :: true?]
            [d (process-pop! ctx "d_2") :: true?])
        (connected? b 0 d 0) :: true?)
      (process-pop! ctx) ==> #f))
  )

(define-test "built-in types (equality)"

  (let ([proc (sexp->process '(("foo" ("a" ("b" ("c"))) ("a" ("c" ("b"))))))]
        [lreg (make-register 2)]
        [preg (make-register 2)])
    ((rule-match-tree '("foo" #f #f)) proc lreg preg #f)
    ((rule-traverse-contexts '(("===" 0 1))) proc lreg preg (default-env))
    ==> #f)

  (let ([proc (sexp->process '(("foo" ("a" ("b" ("c"))) ("a" ("b" ("c"))))))]
        [lreg (make-register 2)]
        [preg (make-register 2)])
    ((rule-match-tree '("foo" #f #f)) proc lreg preg #f)
    ((rule-traverse-contexts '(("===" 0 1))) proc lreg preg (default-env))
    :: true?
    (let* ([res (preg 1)]
           [foo (process-get proc "foo_2")]
           [as (res "a_2") :: (^l (= (length l) 2))])
      (length (res "b_2")) ==> 2
      (length (res "c_1")) ==> 2
      (res 0) ==> (foo '(0))
      (res 1) ==> (foo '(1))))
  )

;;   + | user-defined types

;; **FIXME** ANNOYING
(define (type-user-defined proc preg type type-env :rest ports)
  ;;  Search for a subprocess of PROC whose ports are exactly PORTS
  ;; and satisfies user-defined TYPE wrt TYPE-ENV. If any, return the
  ;; process. Otherwise return #f. Atoms in PREG will not match.
  (let* ([proc (rlet1 proc (process-copy proc) ;; process of remaining atoms
                 (map-register preg (cute map-atoms <> (pa$ process-pop! proc))))]
         [arity (length ports)]
         [res (make-process arity)])
    (let set-ports ([port-nth-pairs (map cons ports (iota arity))])
      (when (pair? port-nth-pairs)
        (let ([port (caar port-nth-pairs)]
              [n (cdar port-nth-pairs)])
          (cond [(assoc (partner port) port-nth-pairs)
                 => (^p (alist-delete! (car p) port-nth-pairs)
                        (connect res (list n) res (list (cdr p))))]
                [else
                 (set! (res (list n)) port)]))
        (set-ports (cdr port-nth-pairs))))
    (let/cc return
      (dolist (clause (type 'clauses))
        (let ([lreg (make-register (car (clause 'spec)))]
              [preg (make-register (car (clause 'spec)))])
          (dolist (port ports) (register-push! lreg port))
          (when ((apply >> (append (map rule-match-tree (clause 'trees))
                                   (list (rule-check-links (clause 'links))
                                         (rule-traverse-contexts (clause 'goals))
                                         (rule-delete-processes!))))
                 proc lreg preg type-env)
            (map-register preg (cute map-atoms <> (pa$ process-push! res)))
            (return res))))
      #f)))

(define-test "user-defined types"
  (let* ([type (make <type>
                 ;; type list(X) { nil(X). cons(H, T, X) :- ground(H), list(T). }
                 :clauses (list (make <type-clause>
                                  :spec '(2 . 1)
                                  :trees '(("nil" #f))
                                  :links '((0 . 1)))
                                (make <type-clause>
                                  :spec '(4 . 3)
                                  :trees '(("cons" #f #f #f))
                                  :links '((0 . 3))
                                  :goals '(("ground" 1) ("list" 2)))))]
         [type-env (hash-table 'string=?
                               (cons "list" type)
                               (cons "ground" type-subr-ground))])
    ;; tests
    (let1 proc (sexp->process '(("foo" ("nil"))))
      ((>> (rule-match-tree '("foo" #f))
           (rule-traverse-contexts '(("list" 0))))
       proc (make-register 1) (make-register 2) type-env) :: true?)
    (let1 proc (sexp->process
                '(("foo" ("cons" ("a") ("cons" ("b") ("cons" ("c") ("nil")))))))
      ((>> (rule-match-tree '("foo" #f))
           (rule-traverse-contexts '(("list" 0))))
       proc (make-register 1) (make-register 2) type-env) :: true?)
    (let1 proc (sexp->process
                '(("foo" ("cons" ("a") ("cons" ("b") ("cons" ("c") ("d")))))))
      ((>> (rule-match-tree '("foo" #f))
           (rule-traverse-contexts '(("list" 0))))
       proc (make-register 1) (make-register 2) type-env) ==> #f)
    ))

;;   + | split

(define (rule-split-contexts! context)
  (^(proc lreg preg type-env :optional next)
    ;;  Extract grounds from context and push them to PREG. If
    ;; impossible, return #f. CONTEXT may be destructed.
    ;;  If NEXT is specified, return value for success is
    ;; (next). Otherwise a boolean value.
    (let ([preg-state (register-get-state preg)]
          [preg2 (make-register (length (cdr context)))]
          [process (preg (car context))])
      (dump +1 ">> split contexts " (process->sexp process) ".")
      (if (and ((>> (rule-traverse-contexts
                     (map (pa$ cons "ground") (cdr context)))
                    (rule-delete-processes!)) process lreg preg2 (default-env))
               (zero? (length (map-atoms process (^a #t)))))
          (begin (dump -1 "<< splitted to " (map-register preg2 process->sexp) ".")
                 (rlet1 res (begin (map-register preg2 (pa$ register-push! preg))
                                   (or (undefined? next) (next)))
                   (unless res (register-restore-state! preg preg-state))))
          (begin (dump -1 "XX split failed.")
                 #f)))))

(define-test "rule-split-contexts"

  (let ([proc (sexp->process '(("foo" ("a" ("b" ("c"))) ("a" ("b" ("c"))))))]
        [lreg (make-register 2)]
        [preg (make-register 4)])
    ((>> (rule-match-tree '("foo" #f #f))
         (rule-traverse-contexts '(("===" 0 1)))
         (rule-split-contexts! '(1 (0) (1))))
     proc lreg preg (default-env)) :: true?
     (process->sexp (preg 2)) ==> '(("a" ("b" ("c")) 0))
     (process->sexp (preg 3)) ==> '(("a" ("b" ("c")) 0)))

  (let ([proc (sexp->process '(("foo" ("a" ("b" L1)) ("a" ("b" L1)))))]
        [lreg (make-register 2)]
        [preg (make-register 4)])
    ((>> (rule-match-tree '("foo" #f #f))
         (rule-traverse-contexts '(("ground" 0 1)))
         (rule-split-contexts! '(1 (0) (1))))
     proc lreg preg (default-env)) ==> #f)
  )

;;   + | body

;; **FIXME** "NEXT" CANNOT FAIL (maybe no problem)
(define (rule-instantiate-body! trees)
  (^(proc lreg preg type-env :optional next)
    ;;  Instantiate TREES wrt LREG and PREG, into PROC. When optional
    ;; argument next is specified, it is called with no arguments at
    ;; last.
    (define (instantiate-tree tree :optional (parent #f))
      (if (pair? tree)
          (cond [(string? (car tree)) ;; atom
                 (let* ([args (cdr tree)]
                        [children (length args)]
                        [arity (+ children (if parent 1 0))]
                        [atom (make-atom (car tree) arity)])
                   (dump 0 "new atom " (atom 'functor) ".")
                   (process-push! proc atom)
                   (when parent
                     (connect atom (- arity 1) (parent 'atom) (parent 'nth)))
                   (dotimes (n children)
                     (instantiate-tree (car args) (atom (list n)))
                     (set! args (cdr args))))]
                [(integer? (car tree)) ;; process
                 (let* ([newproc (process-deep-copy (preg (car tree)))]
                        [arity (newproc 'arity)]
                        [children (- arity (if parent 1 0))]
                        [args (cdr tree)])
                   (unless (= (length args) children)
                     (error "rule-instantiate-body!: arity does not match."))
                   (dump 0 "new process " (process->sexp newproc) ".")
                   (map-atoms newproc (pa$ process-push! proc))
                   (when parent
                     (connect newproc (- arity 1) (parent 'atom) (parent 'nth)))
                   (dotimes (n children)
                     (instantiate-tree (car args) (newproc (list n)))
                     (set! args (cdr args))))])
          (cond [(not tree) ;; new link
                 (register-push! lreg parent)]
                [(integer? tree) ;; known link
                 (let1 partner (lreg tree)
                   (connect (parent 'atom) (parent 'nth)
                            (partner 'atom) (partner 'nth)))])))
    (dump +1 ">> instantiate body.")
    (map (^t (instantiate-tree t)) trees)
    (dump -1 "<< instantiate done.")
    (or (undefined? next) (next))))

(define-test "instantiate-body"

  ;; simple rewriting
  (let1 proc (sexp->process
              ;; foo(x, y) foo(a, b).
              '(("foo" ("x") ("y")) ("foo" ("a") ("b"))))
    ;; foo(L1, L2) :- 1(L1), 2(L2).
    (let1 count 0 ;;                       l0  l1
      (while ((>> (rule-match-tree '("foo" #f #f)) ;; -> p0
                  (rule-instantiate-body! '(("1" 0) ("2" 1)))
                  (rule-delete-processes!))
              proc (make-register 2) (make-register 1) (default-env))
        (inc! count))
      count ==> 2)
    ;; 1(x), 1(a), 2(y), 2(b)
    ((((process-pop! proc "x_1") 0) 'atom) 'name) ==> "1"
    ((((process-pop! proc "y_1") 0) 'atom) 'name) ==> "2"
    ((((process-pop! proc "a_1") 0) 'atom) 'name) ==> "1"
    ((((process-pop! proc "b_1") 0) 'atom) 'name) ==> "2"
    (length (map-atoms proc identity)) ==> 4)

  ;; copying / removing contexts
  (let1 proc (sexp->process
              ;; foo(L1, x(L1, y)), foo(L2, L3), bar(L2, L3)
              '(("foo" L1 ("x" L1 ("y"))) ("foo" L2 L3) ("bar" L2 L3)))
    ;; foo(L1, L2), foo(L3, L4), $x[L1, L2], $y[L3, L4]
    ;;     :- atom($x) | foo(LL1, LL2), $x[LL1, LL2], foo(LL3, LL4), $x[LL3, LL4]
    ((>> (rule-match-tree '("foo" #f #f)) ;; -> p0
         (rule-match-tree '("foo" #f #f)) ;; -> p1
         (rule-traverse-contexts '(("atom" 0 1)
                                   ("ground" 2 3))) ;; -> p2, p3
         (rule-instantiate-body! '(("foo" #f (3 4)) ("foo" #f (3 5))))
         (rule-delete-processes!))
     proc (make-register 6) (make-register 4) (default-env))
    :: true?
    ;; foo(L1, x(L1, y)), foo(L2, x(L2, y))
    (let1 fn (^p (cons ((p 'atom) 'name) (p 'nth)))
      (map-args (process-pop! proc "foo_2") fn) ==> '(("x" . 0) ("x" . 2))
      (map-args (process-pop! proc "foo_2") fn) ==> '(("x" . 0) ("x" . 2))
      (map-args (process-pop! proc "x_3") fn) ==> '(("foo" . 0) ("y" . 0) ("foo" . 1))
      (map-args (process-pop! proc "x_3") fn) ==> '(("foo" . 0) ("y" . 0) ("foo" . 1)))
    (map-atoms proc (^a (a 'name))) ==> '("y" "y"))

  ;; process with direct links
  (let1 proc (sexp->process
              ;; foo(L1, L2, L3), bar(L2, L1, a(L3))
              '(("foo" L1 L2 L3) ("bar" L2 L1 ("a" L3))))
    ;; foo(L1, L2, L3), bar(L4, L5, L6), $x[L1, L2, L3, L4, L5, L6] :- $x[a, b, x, c, d, y].
    ;;      +-foo---+           a  b    x
    ;;      |  |    |           |  |    |
    ;;  /---+--+----+-\     /---+--+----+-\
    ;;  |   |  |    | |     |   |  |    | |
    ;;  | +-+--+    | |     | +-+--+    | |
    ;;  | | |       | |     | | |       | |
    ;;  | | +--+    a | ==> | | +--+    a |
    ;;  | |    |    | |     | |    |    | |
    ;;  \-+----+----+-/     \-+----+----+-/
    ;;    |    |    |         |    |    |
    ;;    +---bar---+         c    d    y
    ((>> (rule-match-tree '("foo" #f #f #f)) ;; -> p0
         (rule-match-tree '("bar" #f #f #f)) ;; -> p1
         (rule-traverse-contexts '(("ground" 0 1 2 3 4 5))) ;; -> p2
         (rule-instantiate-body! '((2 ("a") ("b") ("x") ("c") ("d") ("y"))))
         (rule-delete-processes!))
     proc (make-register 6) (make-register 3) (default-env))
    :: true?
    ;; a(d), b(c), y(a(x))
    ((((process-pop! proc "a_1") 0) 'atom) 'name) ==> "d"
    ((((process-pop! proc "b_1") 0) 'atom) 'name) ==> "c"
    ((((process-pop! proc "y_1") 0) 'atom) 'name) ==> "a"
    ((((process-pop! proc "x_1") 0) 'atom) 'name) ==> "a"
    (map-args (process-pop! proc "a_2") (^p ((p 'atom) 'name))) ==> '("x" "y")
    (map-atoms proc (^a (a 'arity))) ==> '(1 1))
  )

(define-test "backtracking with or>"
  (let1 proc (sexp->process
              ;; foo(x, y) foo(a, b).
              '(("foo" ("1")) ("foo" ("a")) ("foo" ("a" ("b")))))
    ;; foo($x) :- { int($x) | foo(num).
    ;;                atom($x) | foo(atm).
    ;;              ground($x) | foo(gnd). }
    (while ((>> (rule-match-tree '("foo" #f))
                (>or (>> (rule-traverse-contexts '(("int" 0)))
                         (rule-instantiate-body! '(("bar" ("num")))))
                     (>> (rule-traverse-contexts '(("atom" 0)))
                         (rule-instantiate-body! '(("bar" ("atm")))))
                     (>> (rule-traverse-contexts '(("ground" 0)))
                         (rule-instantiate-body! '(("bar" ("gnd"))))))
                (rule-delete-processes!))
            proc (make-register 1) (make-register 2) (default-env)))
    ;; foo(num), foo(atm), foo(gnd).
    (process-pop! proc "num_1") :: true?
    (process-pop! proc "atm_1") :: true?
    (process-pop! proc "gnd_1") :: true?
    (map-atoms proc (^a (a 'name))) ==> '("bar" "bar" "bar")))

;;   + | apply rule

(define (apply-rule! process rule type-env)
  (when *debug*
    (print (if (procedure? rule) rule (slot-ref rule 'trees)))
    (flush))
  (cond [(is-a? rule <rule>)
         ((apply >> (append
                     (map rule-match-tree (rule 'trees))
                     (list (rule-check-links (rule 'links))
                           ($ apply >or
                              $ map (^c ($ apply >> (rule-traverse-contexts (c 'contexts))
                                           $ append (map rule-split-contexts! (c 'split))
                                           $ list (rule-instantiate-body! (c 'body))))
                              $ rule 'clauses)
                           (rule-delete-processes!))))
          process
          (make-register (car (rule 'spec)))
          (make-register (cdr (rule 'spec)))
          type-env)]
        [(procedure? rule)
         (rule process)]
        [else
         (error "apply-rule!: invalid rule.")]))

(define-test "apply-rule!"

  ;; foo(L1, a, b, L1).
  ;; foo(L1, $x, $y, L1) :- atom($x), atom($y) | bar($x, $y)
  (let ([process (sexp->process '(("foo" L1 ("a") ("b") L1)))]
        [rule (make <rule>
                :spec '(4 . 3)
                :trees '(("foo" #f #f #f #f))
                :links '((0 . 3))
                :clauses `(,(make <rule-clause>
                              :contexts '(("atom" 1) ("atom" 2))
                              :body '(("bar" (1) (2))))))])
    (apply-rule! process rule (default-env)) :: true?
    (process->sexp process) ==> '(("bar" ("a") ("b")))
    (apply-rule! process rule (default-env)) ==> #f)

  ;; foo(a(b)), foo(1), foo(a).
  ;; foo($x) :- { int($x) | bar(num).
  ;;                atom($x) | bar(atm).
  ;;              ground($x) | bar(gnd). }
  (let ([process (sexp->process '(("foo" ("a" ("b"))) ("foo" ("1")) ("foo" ("a"))))]
        [rule (make <rule>
                :spec '(1 . 2)
                :trees '(("foo" #f))
                :clauses (list (make <rule-clause>
                                 :contexts '(("int" 0))
                                 :body '(("bar" ("num"))))
                               (make <rule-clause>
                                 :contexts '(("atom" 0))
                                 :body '(("bar" ("atm"))))
                               (make <rule-clause>
                                 :contexts '(("ground" 0))
                                 :body '(("bar" ("gnd"))))))])
    ;; bar(num), bar(atm), bar(gnd).
    (apply-rule! process rule (default-env)) :: true?
    (apply-rule! process rule (default-env)) :: true?
    (apply-rule! process rule (default-env)) :: true?
    (apply-rule! process rule (default-env)) ==> #f
    (process-pop! process "num_1") :: true?
    (process-pop! process "atm_1") :: true?
    (process-pop! process "gnd_1") :: true?
    (map-atoms process (^a (a 'name))) ==> '("bar" "bar" "bar"))

  (let ([process (sexp->process '(("foo" ("a" ("b")) ("a" ("b")))))]
        [rule (make <rule>
                :spec '(2 . 4)
                :trees '(("foo" #f #f))
                :clauses (list (make <rule-clause>
                                 :contexts '(("===" 0 1))
                                 :split '((1 (0) (1)))
                                 :body '(("bar" (2)) ("bar" (3))))))])
    (apply-rule! process rule (default-env)) :: true?
    (process->sexp process) ==> '(("bar" ("a" ("b"))) ("bar" ("a" ("b"))))
    (apply-rule! process rule (default-env)) ==> #f)
  )

;;   + | built-in rules

(define (rule-subr-connect! proc)
  ;;  Process all occurrence of "=_2" in PROC.
  (if-let1 atom (process-pop! proc "=_2")
    (let ([port1 (atom 0)] [port2 (atom 1)])
      (connect (port1 'atom) (port1 'nth)  (port2 'atom) (port2 'nth))
      #t)
    #f))

(define (rule-subr-gensym! proc)
  ;;  Process all occurrence of "gensym_1" in PROC.
  (and-let* ([atom (process-pop! proc "gensym_1")]
             [newatom (make-atom (symbol->string (gensym "g")) 1)])
    (process-push! proc newatom)
    (connect newatom 0 ((atom 0) 'atom) 0)
    #t))

(define (rule-subr-arithmetic-binary! fn name proc)
  ;;  Process all occurrence of binary (arithmetic) operation NAME in
  ;; PROC with FN.
  (and-let* ([atom (process-get proc #`",|name|_3")]
             [latom ((atom 0) 'atom)]
             [ratom ((atom 1) 'atom)]
             [_ (= (latom 'arity) 1)]
             [_ (= (ratom 'arity) 1)]
             [lval (string->number (latom 'name))]
             [_ (integer? lval)]
             [rval (string->number (ratom 'name))]
             [_ (integer? rval)]
             [return (atom 2)]
             [newatom (make-atom (number->string (fn lval rval)) 1)])
    (process-pop! proc atom)
    (process-pop! proc latom)
    (process-pop! proc ratom)
    (process-push! proc newatom)
    (connect (return 'atom) (return 'nth) newatom 0)
    #t))
(define rule-subr-add! (pa$ rule-subr-arithmetic-binary! + "+"))
(define rule-subr-sub! (pa$ rule-subr-arithmetic-binary! - "-"))
(define rule-subr-mul! (pa$ rule-subr-arithmetic-binary! * "*"))
(define rule-subr-div! (pa$ rule-subr-arithmetic-binary! quotient "/"))
(define rule-subr-mod! (pa$ rule-subr-arithmetic-binary! modulo "mod"))

(define (rule-subr-arithmetic-unary! fn name proc)
  ;;  Process all occurrence of unary (arithmetic) operation NAME in
  ;; PROC with FN.
  (and-let* ([atom (process-get proc #`",|name|_2")]
             [opatom ((atom 0) 'atom)]
             [_ (= (opatom 'arity) 1)]
             [opval (string->number (opatom 'name))]
             [_ (integer? opval)]
             [return (atom 1)]
             [newatom (make-atom (number->string (fn opval)) 1)])
    (process-pop! proc atom)
    (process-pop! proc opatom)
    (process-push! proc newatom)
    (connect (return 'atom) (return 'nth) newatom 0)
    #t))
(define rule-subr-pos! (pa$ rule-subr-arithmetic-unary! + "+"))
(define rule-subr-neg! (pa$ rule-subr-arithmetic-unary! - "-"))

(define (default-system)
  (list rule-subr-connect!
        rule-subr-gensym!
        rule-subr-add!
        rule-subr-sub!
        rule-subr-mul!
        rule-subr-div!
        rule-subr-mod!
        rule-subr-pos!
        rule-subr-neg!))

(define-test "built-in operations"
  (let1 proc (sexp->process
              '(("bar" ("+" ("1")) ("-" ("2"))
                 ("+" ("1") ("2")) ("-" ("3") ("4"))
                 ("*" ("5") ("6")) ("/" ("7") ("8")))))
    (while (any (cute apply-rule! proc <> (default-env)) (default-system)))
    (map-args (process-pop! proc "bar_6") (^p ((p 'atom) 'name)))
    ==> '("1" "-2" "3" "-1" "30" "0")))

;; + parser / unparser
;;   + parser

(define $comment
  ($or ($seq ($or ($c #\%) ($s "//"))
             ($many ($none-of #[\n]))
             ($or ($c #\newline) $eof))
       ($seq ($s "/*")
             ($many-till ($any) ($s "*/")))))

(define $spaces
  ($many ($or $comment ($any #[\x09-\x0d ]))))

(define ($token parser :optional [name #f])
  (if name ($g ($token parser)) ($seq0 parser ($skip $spaces))))

(define (make-expr-parser term ops)
  ;;  Make a parser that parses a pre/in-fix expression. OPS must be a
  ;; list of operators of the form (PRIORITY TYPE OP OP ...) where
  ;; PRIORITY is an integer and TYPE is one of 'left 'right or
  ;; 'prefix. OP is a list of the form (STR FN) for infix operators or
  ;; (STR FN . ARITY) for prefix operators, where ARITY is a natural
  ;; number or nil. If ARITY is nil, the operator is counted as an
  ;; unray operator.
  (rec expr
    (let loop ([ops (sort ops (^(l1 l2) (< (car l1) (car l2))))])
      (if (null? ops)
          ($or ($between ($token ($c #\())
                         ($g ($delay expr) "expression")
                         ($token ($c #\))))
               term)
          (let ([$subexp ($g (loop (cdr ops)) "term")]
                [$op ($ apply $or
                        $ map (^o ($seq ($s (car o)) ($ret (cdr o))))
                        $ sort (cddar ops) (^(o1 o2) (string> (car o1) (car o2))))])
            (case (cadar ops)
              [(prefix)
               (rec self
                 ($or ($g ($do* op <- ($token $op)
                                terms <- ($rep ($delay self)
                                               (if (integer? (cdr op)) (cdr op) 1))
                                ($ret (apply (car op) terms)))
                          "prefix operation" #t)
                      $subexp))]
              [(left)
               ($chain-left $subexp ($token ($<< car $op)))]
              [(right)
               ($chain-right $subexp ($token ($<< car $op)))]))))))

(define-test "expression parser"
  (letrec ([nat ($do* x <- ($token ($many1 ($any #[0-9])))
                      (let1 val (string->number (list->string x))
                        ($ret val #`"integer literal ',x'")))]
           [list ($between ($token ($c #\[))
                           ($g ($sep-by ($delay parser) ($token ($c #\,)))
                               "list elements")
                           ($token ($c #\])))]
           [parser (make-expr-parser
                    ($or nat list)
                    `((0    left   ("times" ,*))
                      (1    prefix ("twice" ,(pa$ * 2)) ("multiply" ,* . 2))
                      (100  left   ("*" ,*) ("/" ,/))
                      (200  right  ("^" ,expt))
                      (50   left   ("-" ,-) ("+" ,+) ("++" ,append!))
                      (1000 prefix ("-" ,-) ("+" ,+) ("+/" ,(pa$ apply +)))))])
    (parse-string parser "1 / 2 + 3 * 4 - 5") ==> 15/2
    (parse-string parser "1 / ( 2 + 3 ) * 4 - 5") ==> -21/5
    (parse-string parser "+/ ( [1 , 2, 3 , 4, 5 ] ++ [6, 7 , 8] ) * 6") ==> 216
    (parse-string parser "- +/[+/[ 2], 1 * 2 ] * 2 ^ 2 ^ 2") ==> -64
    (parse-string parser "- - 1") ==> 1
    (parse-string parser "multiply 2 twice 1 + 2 * 3 times 10") ==> 280))

;; Parser<String>
(define $atomname
  ($token
   ($do* x <- ($<< list->string
                   ($or ($many1 ($any #[0-9]))
                        ($cons ($any #[a-z_]) ($many ($any #[a-zA-Z0-9_])))
                        ($between ($c #\') ($many1 ($none-of #['])) ($c #\'))))
         ($ret x #`"atom name ',x'"))
   "atom name"))

;; Parser<Symbol>
(define $contextname
  ($token ($do* x <- ($<< list->string
                          ($cons ($c #\$) ($many1 ($any #[a-zA-Z0-9_]))))
                ($ret (string->symbol x) #`"context name ',x'"))
          "context name"))

;; Parser<Symbol>
(define $link
  ($token ($do* x <- ($<< list->string
                          ($cons ($any #[A-Z]) ($many ($any #[a-zA-Z0-9_]))))
                ($ret (string->symbol x) #`"link ',x'"))
          "link"))

;; Parser<Expr>
(define ($expr ops)
  (letrec ([exprs ($sep-by ($delay expr) ($token ($c #\,)))]
           [term ($or $link
                      ($cons $atomname
                             ($optional
                              ($between ($token ($c #\()) exprs ($token ($c #\))))
                              ()))
                      ($cons $contextname
                             ($optional
                              ($between ($token ($c #\[)) exprs ($token ($c #\])))
                              ()))
                      ($between ($token ($c #\[))
                                ($do* init <- exprs
                                      last <- ($optional
                                               ($seq ($token ($c #\|)) ($delay expr))
                                               (list "[]"))
                                      ($ret (fold-right (pa$ list ".") last init)))
                                ($token ($c #\]))))]
           [expr (make-expr-parser term ops)])
    expr))

(define (default-ops)
  `((1000 right
          ("=" ,(^(x y) (list "=" x y))))
    (2000 right
          ("===" ,(^(x y) (list "===" x y)))
          ("==" ,(^(x y) (list "==" x y)))
          (">=" ,(^(x y) (list ">=" x y)))
          ("<=" ,(^(x y) (list "<=" x y)))
          ("<" ,(^(x y) (list "<" x y)))
          (">" ,(^(x y) (list ">" x y))))
    (3000 left
          ("+" ,(^(x y) (list "+" x y)))
          ("-" ,(^(x y) (list "-" x y))))
    (4000 left
          ("*" ,(^(x y) (list "*" x y)))
          ("/" ,(^(x y) (list "/" x y)))
          ("mod" ,(^(x y) (list "mod" x y))))
    (9000 prefix
          ("-" ,(^x (list "-" x)) . 1))))

(define-test "parse $trees"
  (let1 parser ($sep-by ($expr (default-ops)) ($token ($c #\,)))
    (parse-string parser "a(b(c), d, L1, L2), e(L1, L2)")
    ==> '(("a" ("b" ("c")) ("d") L1 L2) ("e" L1 L2))
    (parse-string parser "'a long name atom'('a long name child')")
    ==> '(("a long name atom" ("a long name child")))
    (parse-string parser "a(b, X), c(Y), $x[X, Y]")
    ==> '(("a" ("b") X) ("c" Y) ($x X Y))
    (parse-string parser "a(b == c == ( d + 4 ) * 10 * e(f))")
    ==> '(("a" ("==" ("b") ("==" ("c") ("*" ("*" ("+" ("d") ("4")) ("10")) ("e" ("f")))))))
    (parse-string parser "[], [a(b)+c, d], [e(f)|G]")
    ==> '(("[]") ("." ("+" ("a" ("b")) ("c")) ("." ("d") ("[]"))) ("." ("e" ("f")) G))))

(define ($statement ops)
  (letrec ([block (^p ($or ($between ($token ($c #\{))
                                     ($followed-by p ($token ($c #\.)))
                                     ($c #\}))
                           ($do x <- p
                                ($c #\.)
                                ($ret (list x)))))]
           [expr ($expr ops)]
           [exprs ($sep-by expr ($token ($c #\,)))]
           [rhs (block ($do y <- exprs
                            z <- ($optional ($seq ($token ($c #\|)) exprs) #f)
                            ($ret (if z (cons y z) (cons () y)))))]
           [process ($do l <- exprs
                         r <- ($or ($seq ($c #\.) ($ret #f))
                                   ($seq ($token ($s ":-")) rhs))
                         ($ret (if r (cons* 'rule l r) (cons 'process l))))]
           [sysrule ($do ($token ($s "define"))
                         l <- exprs
                         ($token ($s ":-"))
                         r <- rhs
                         ($ret (cons* 'sysrule l r)))]
           [typedef ($do ($token ($s "typedef"))
                         type <- expr
                         body <- (block ($do lhs <- exprs
                                             rhs <- ($optional
                                                     ($seq ($token ($s ":-")) exprs)
                                                     ())
                                             ($ret (cons lhs rhs))))
                         ($ret (cons* 'typedef type body)))]
           [defop ($do ($token ($s "defop"))
                       x <- exprs
                       ($c #\.)
                       ($ret (cons 'defop x)))]
           [command ($do ($c #\:)
                         x <- ($or ($s "quit")
                                   ($s "clear")
                                   ($s "reset")
                                   ($s "trace")
                                   ($s "debug")
                                   ($s "interactive"))
                         ($ret (list 'command (string->symbol x))))])
    ($seq ($skip $spaces)
          ($or ($g process "process" #t) sysrule typedef defop command))))

(define-test "parse $statement"
  (let1 parser ($statement (default-ops))
    (parse-string parser "defop infix, '=', 100, right.")
    ==> '(defop ("infix") ("=") ("100") ("right"))
    (parse-string parser "typedef list(X) { X=[]. X=[H, T] :- list(T). }")
    ==> '(typedef ("list" X)
                  ((("=" X ("[]"))) .  ())
                  ((("=" X ("." H ("." T ("[]"))))) . (("list" T))))
    (parse-string parser "foo(X), bar(baz(Y), X), foobar(Y).")
    ==> '(process ("foo" X) ("bar" ("baz" Y) X) ("foobar" Y))
    (parse-string parser "foo(X), bar(baz(Y), X) :- foobar(Y).")
    ==> '(rule (("foo" X) ("bar" ("baz" Y) X))
               (() . (("foobar" Y))))
    (parse-string parser "foo(X), $x[X] :- ground($x) | foobar(X).")
    ==> '(rule (("foo" X) ($x X))
               ((("ground" ($x))) . (("foobar" X))))
    (parse-string parser "foo(X), $x[X] :- {ground($x)|gnd(X). atom($x)|atm(X).}")
    ==> '(rule (("foo" X) ($x X))
               ((("ground" ($x))) . (("gnd" X)))
               ((("atom" ($x))) . (("atm" X))))
    (parse-string parser " :- unique | .")
    ==> '(rule ()
               ((("unique")) . ()))
    (parse-string parser "define append([], X, Ret) :- Ret=X .")
    ==> '(sysrule (("append" ("[]") X Ret))
                  (() . (("=" Ret X))))
    (parse-string parser "define , foo :- bar .")
    ==> '(rule
          (("define") ("foo"))
          (() . (("bar"))))
    (parse-string parser ":quit")
    ==> '(command quit)))

(define (read-statement ops)
  ;; Read a statement from (current-input-port).
  (parse-port ($statement ops)))

;;   + unparser

(define (process->string proc)
  (define (trees->string trees)
    (define (tree->string tree)
      (cond [(symbol? tree) (symbol->string tree)]
            ;; has 1 or more arguments
            [(pair? (cdr tree)) #`",(car tree)(,(trees->string (cdr tree)))"]
            ;; has no arguments
            [else #`",(car tree)"]))
    (if (null? trees) ""
        (string-append
         (tree->string (car trees))
         (string-concatenate (map (^t #`"\, ,(tree->string t)") (cdr trees))))))
  (string-append
   (trees->string (process->sexp proc)) "."))

;; + compiler
;;   + rule compiler

;; **FIXME** POOR IMPLEMENTATION
(define (compile-rule rule)
  ;;  Compile rule ast to a <rule> object.
  (let ([lcount 0]
        [pcount 0]
        [trees ()]
        [contexts ()]
        [links ()]
        [freelinks ()])

    (define (eliminate-=s)
      ;;  '(("=" ("foo") L1) ("=" L2 ("bar")) ("=" L1 L2))
      ;; => '(("foo" L1) ("bar" L2) ($x L1 L1)), $x::linked
      (set-car! rule
                (map (^t (or (and (pair? t)
                                  (string? (car t))
                                  (string=? (car t) "=")
                                  (cond [(pair? (cadr t)) ;; a = L1 / a = b
                                         (append! (cadr t) (cddr t))]
                                        [(pair? (caddr t)) ;; L1 = a
                                         (append! (caddr t) (list (cadr t)))]
                                        [else ;; L1 = L2
                                         (let1 newsymbol (gensym "$'")
                                           (map (^c (push! (car c)
                                                           (list "linked" (list newsymbol))))
                                                (cdr rule))
                                           (list newsymbol (cadr t) (caddr t)))]))
                             t))
                     (car rule))))

    (define (extract-template)
      ;;  '(($y X) ("foo" ($x ("a" Y Y)) X) ("bar" Z))
      ;; -- (extract-template) --
      ;;      trees : (("foo" L1 L2) ("a" L3 L3 L4) ("bar" L5))
      ;;   contexts : (($y . (L2)) ($x . (L4 L1)))
      ;;  freelinks : ((Z . L5))
      (map
       (rec (extract-tree tree :optional (root #t) (slice #f))
         (if (pair? tree)
             (cond [(symbol? (car tree)) ;; context
                    (when (assq (car tree) contexts)
                      (lmn-error #`"each processes must appear exactly once in the LHS."))
                    (let1 children (map-in-order (cute extract-tree <> #f #t)
                                                 (cdr tree))
                      (if root
                          (push! contexts (cons (car tree) children))
                          (rlet1 newsymbol (gensym "L")
                            (push! contexts (cons (car tree)
                                                  (append! children (list newsymbol)))))))]
                   [(string? (car tree)) ;; atom
                    (let1 res (cons (car tree)
                                    (map-in-order (cute extract-tree <> #f #f)
                                                  (cdr tree)))
                      (cond [root (push! trees res)]
                            [slice (rlet1 newsymbol (gensym "L")
                                     (push! trees (append! res (list newsymbol))))]
                            [else res]))])
             (if-let1 link (assq-ref freelinks tree) ;; link
               (begin (set! freelinks (alist-delete! tree freelinks)) link)
               (rlet1 newsymbol (gensym "L")
                 (push! freelinks (cons tree newsymbol))))))
       (car rule))
      (set! trees (reverse! trees)))

    (define (assign-ports)
      ;;  Check if all contexts have their partner atoms. And assign
      ;; ports to freelinks/contexts so that we may use
      ;; (process->sexp (sexp->process <>)) to reduce the number of trees.
      ;;      trees : (("foo" L1 L2) ("a" L3 L3 L4) ("bar" L5))
      ;;   contexts : (($y L2) ($x L4 L1))
      ;;  freelinks : ((Z . L5))
      ;; -- (assign-ports) --
      ;;      trees : (("foo" 0 1) ("a" L3 L3 2) ("bar" 3))
      ;;   contexts : (($y 1) ($x 2 0))
      ;;  freelinks : ((Z . 3))
      (let* ([count 0]
             [links (apply append (map cdr freelinks) (map cdr contexts))]
             [dictionary ()]
             [newsymbol (^l (begin0 count
                                    (push! dictionary (cons l count))
                                    (inc! count)))]
             [lookup (^l (or (rlet1 newlink (assq-ref dictionary l)
                               (set! dictionary (alist-delete! l dictionary)))
                             l))])
        (map (^c (set-cdr! c (map newsymbol (cdr c)))) contexts)
        (map (^l (set-cdr! l (newsymbol (cdr l)))) freelinks)
        (set! trees
              (map (rec (assign-ports-1 tree)
                     (cond [(pair? tree)
                            (cons (car tree) (map assign-ports-1 (cdr tree)))]
                           [else
                            (lookup tree)]))
                   trees))
        (when (pair? dictionary)
          ;; there're some contexts that do not appear in trees
          (lmn-error #`"each contexts must have at least one partner atom."))))

    (define (compile-head)
      ;;      trees : (("foo" 0 1) ("a" L3 L3 2) ("bar" 3))
      ;;   contexts : (($y 1) ($x 2 0))
      ;;  freelinks : ((Z . 3))
      ;; -- (compile-head) --
      ;;     lcount : 6
      ;;     pcount : 3
      ;;      trees : (("foo" #f #f) ("a" #f #f #f) ("bar" #f))
      ;;   contexts : (($y 1) ($x 4 0))
      ;;      links : ((2 . 3))
      ;;  freelinks : ((Z . 5))
      (set! pcount (length trees))
      (let* ([dictionary ()]
             [lookup (^l (or (assq-ref dictionary l) l))])
        (set! trees
              (map (rec (compile-tree tree)
                     (cond [(pair? tree) ;; atom
                            (cons (car tree)
                                  (map compile-tree (cdr tree)))]
                           [else ;; link
                            (if-let1 partner (assq-ref dictionary tree)
                              (push! links (cons partner lcount))
                              (push! dictionary (cons tree lcount)))
                            (inc! lcount)
                            #f]))
                   trees))
        (map (^c (set-cdr! c (map lookup (cdr c)))) contexts)
        (map (^l (set-cdr! l (lookup (cdr l)))) freelinks)))

    (define (compile-clause constraints trees)
      ;;  Returns (lreg-spec preg-spec <rule-clause>).
      (let ([lcount-local lcount]
            [pcount-local pcount]
            [contexts-local ()]
            [guard ()]
            [split ()]
            [body ()])

        ;;        pcount: 1
        ;;      contexts: (($x 0 1) ($y 2 3) ($z 4))
        ;;   constraints: (("==" ($x) ($y)) ("atom" ($z)))
        ;; -- extract constraints --
        ;;        pcount-l: 3
        ;;           guard: (("==" 0 1 2 3) ("atom" 4))
        ;;      contexts-l: (($x 1 0 1) ($y 1 2 3) ($z 2))
        (dolist (constraint constraints)
          (let ([type (car constraint)]
                [preg-index pcount-local])
            (unless type
              (lmn-error #`"unknown type ,(car constraint)."))
            (inc! pcount-local)
            (match (cdr constraint)
              [() ;; 0-ary type
               (push! guard (list type))]
              [((p1)) ;; 1-ary type
               (push! contexts-local (cons p1 preg-index))
               (push! guard (cons type (assq-ref contexts p1)))]
              [((ps) ...) ;; n>1-ary types
               (push! guard
                      (cons type
                            (concatenate
                             (map (^p (rlet1 links (assq-ref contexts p)
                                        (push! contexts-local (cons p (cons preg-index links)))))
                                  ps))))]
              [else
               (lmn-error #`"invalid type constraint: ,|constraint|.")])))
        (unless (= (length contexts-local) (length contexts))
          (lmn-error "each contexts must have exactly one type constraint."))
        (set! contexts-local (reverse! contexts-local))
        (set! guard (reverse! guard))

        ;;     pcount-l: 3
        ;;        guard: (("==" 0 1 2 3) ("atom" 4))
        ;;   contexts-l: (($x 1 0 1) ($y 1 2 3) ($z 2))
        ;; -- split contexts --
        ;;     pcount-l: 5
        ;;        split: ((1 (0 1) (2 3)))
        ;;   contexts-l: (($x 3) ($y 4) ($z 2))
        (map (^c (when (pair? (cdr c))
                   (if-let1 pair (assv (cadr c) split)
                     (set-cdr! pair (cons (cddr c) (cdr pair)))
                     (push! split (list (cadr c) (cddr c))))
                   (set-cdr! c pcount-local)
                   (inc! pcount-local)))
             contexts-local)
        (map (^s (set-cdr! s (reverse! (cdr s)))) split)

        ;;       lcount: 5
        ;;        trees: (("a" ($x L1 L2 L3) L1 L2 L3 L4) ("b" ($y) L4))
        ;;   contexts-l: (($x 3) ($y 4) ($z 2))
        ;; -- process body --
        ;;     lcount-l: 9
        ;;         body: (("a" (3 #f #f #f) 5 6 7 #f) ("b" (4) 8))
        (let1 freelinks-local (list-copy freelinks)
          (set! body
                (map (rec (process-tree tree)
                       tree
                       (if (pair? tree)
                           (cond [(string? (car tree))
                                  (cons (car tree) (map process-tree (cdr tree)))]
                                 [(symbol? (car tree))
                                  (cons (assq-ref contexts-local (car tree))
                                        (map process-tree (cdr tree)))])
                           (if-let1 partner (assq-ref freelinks-local tree)
                             (begin (set! freelinks-local (alist-delete! tree freelinks-local))
                                    partner)
                             (begin (push! freelinks-local (cons tree lcount-local))
                                    (inc! lcount-local)
                                    #f))))
                     trees))
          (when (pair? freelinks-local)
            (lmn-error #`"link without a partner (,(caar freelinks-local)).")))

        ;; make result
        (list lcount-local pcount-local
              (make <rule-clause> :contexts guard :split split :body body))
        ))

    (eliminate-=s)
    (extract-template)
    ;; may reduce number of trees
    (assign-ports)
    (set! trees (process->sexp (sexp->process trees)))
    ;; ----------------------
    (compile-head)
    (let1 compiled-clauses
        (apply zip (map (^c (compile-clause (car c) (cdr c))) (cdr rule)))
      (make <rule>
        :spec (cons (apply max (car compiled-clauses))
                    (apply max (cadr compiled-clauses)))
        :trees trees
        :links links
        :clauses (caddr compiled-clauses)))))

(define-test "compile-rule"
  (let* ([type-env (default-env)]
         [parser ($statement (default-ops))]
         [process1 ((compose sexp->process cdr)
                    (parse-string parser "
a(x(y, X), b), c(x(y, Y), d), e(X, Y, a, b, A, A).
"))]
         [process2 ((compose sexp->process cdr)
                    (parse-string parser "
a(x(y, X), b), c(y(x, Y), d), e(X, Y, b, c, A, A).
"))]
         [rule ((compose compile-rule cdr)
                (parse-string parser "
a($x[X], b), c($y[Y], d), e(X, Y, $z, Z, A, A) :- {
                 $x === $y, atom($z) |  equal($x[X], X, $y[Y], Y), bar($z, Z).
    ground($x), ground($y), atom($z) | nequal($x[X], X, $y[Y], Y), bar($z, Z).
}"))])
    (while (apply-rule! process1 rule type-env))
    (while (apply-rule! process2 rule type-env))
    ;; process1 = ((bar (a) (b)) ( equal (x (y) X) X (x (y) Y) Y))
    ;; process2 = ((bar (b) (c)) (nequal (x (y) X) X (y (x) Y) Y))
    (map-args (process-pop! process1 "bar_2") (^p ((p 'atom) 'functor)))
    ==> '("a_1" "b_1")
    (map-args (process-pop! process2 "bar_2") (^p ((p 'atom) 'functor)))
    ==> '("b_1" "c_1")
    (let1 eq (process-pop! process1 "equal_4")
      eq :: true?
      (((((eq 2) 'atom) 0) 'atom) 'functor) ==> "y_1")
    (let1 neq (process-pop! process2 "nequal_4")
      neq :: true?
      (((((neq 2) 'atom) 0) 'atom) 'functor) ==> "x_1")))

;;   + type compiler

(define (compile-type type)
  ;;  Compile typedef ast to a <type> object.
  (let* ([lcount (length (cdar type))]
         [freelinks (map cons (cdar type) (iota lcount))])

    (define (eliminate-=s clause)
      (let ([head (car clause)]
            [body (cdr clause)])
        (set! head
              ($ delete! ()
                 $ map (^t (if (and (pair? t)
                                    (string=? (car t) "="))
                               (cond [(pair? (cadr t)) ;; a = L1 / a = b
                                      (append! (cadr t) (cddr t))]
                                     [(pair? (caddr t)) ;; L1 = a
                                      (append! (caddr t) (list (cadr t)))]
                                     [else ;; L2 = L1
                                      (push! body (list "linked" (cadr t) (caddr t)))
                                      ()])
                               t))
                 head))
        (cons head body)))

    (define (compile-clause clause)
      (let ([lcount-local lcount]
            [freelinks-local (list-copy freelinks)]
            [head (car clause)]
            [links ()]
            [body (cdr clause)])
        (set! head
              (map (rec (compile-tree tree)
                     (cond [(pair? tree) ;; atom
                            (cons (car tree) (map compile-tree (cdr tree)))]
                           [(assq-ref freelinks-local tree) => ;; known link
                            (^p (set! freelinks-local (alist-delete! tree freelinks-local))
                                (push! links (cons p lcount-local))
                                (inc! lcount-local)
                                #f)]
                           [else ;; unknown link
                            (push! freelinks-local (cons tree lcount-local))
                            (inc! lcount-local)
                            #f]))
                   head))
        (set! body
              (map (^g
                     (cons (car g)
                           (map (^l
                                  (rlet1 port (assoc-ref freelinks-local l)
                                    (unless port
                                      (lmn-error #`"unknown linkname in subgoals: ,|l|."))
                                    (set! freelinks-local (alist-delete! l freelinks-local))))
                                (cdr g))))
                   body))
        (when (pair? freelinks-local)
          (lmn-error #`"link without a partner: ,(caar freelinks-local)."))
        (make <type-clause>
          :spec (cons lcount-local (length head))
          :trees head
          :links links
          :goals body)))

    (make <type> :clauses
          (map (compose compile-clause eliminate-=s)
               (cdr type)))))

;; + interface

(define (repl :optional (debug #f))
  (let/cc break
    (let ([*process* (make-process)]
          [*rules* ()]
          [*system-rules* (default-system)]
          [*types* (default-env)]
          [*ops* (default-ops)]
          [*trace* #t]
          [*interactive* #f]
          [*count* 1])
      (define (dump-trace)
        (when *trace*
          (print "--> " *count* ": " (process->string *process*))
          (flush))
        (inc! *count*))
      (while #t
        (let/cc continue
          (guard (err [(<parse-error> err)
                       (parse-port ($skip ($many ($none-of ;; #[\}.]
                                                           #[#\newline])))
                                   (current-input-port))
                       (print #`"SYNTACTIC ERROR: ,(~ err 'message)")]
                      [(<lmn-exception> err)
                       (print #`"SEMANTIC ERROR: ,(~ err 'message)")]
                      [else
                       (print #`"UNEXPECTED ERROR")
                       (raise err)])
            (display "LMN> ")
            (flush)
            (if-let1 statement (read-statement *ops*)
              (begin
                (when *debug* (print ">>>> " statement))
                (case (car statement)
                  [(rule)
                   (rlet1 rule (compile-rule (cdr statement))
                     (when *debug* (print (rule->string rule)))
                     (push! *rules* rule))]
                  [(sysrule)
                   (rlet1 rule (compile-rule (cdr statement))
                     (print "defined a system rule.")
                     (when *debug* (print (rule->string rule)))
                     ;; connect! must always applied first
                     (append! *system-rules* (list rule)))]
                  [(process)
                   (let1 proc (sexp->process (cdr statement))
                     (process->sexp proc) ;; confirm validity of the process
                     (map-atoms proc (pa$ process-push! *process*))
                     (while (any (cute apply-rule! *process* <> *types*) *system-rules*))
                     (dump-trace))]
                  [(typedef)
                   (let1 type (compile-type (cdr statement))
                     (hash-table-put! *types* (caadr statement) type)
                     (print #`"defined ,(caadr statement).")
                     (when *debug* (print (type->string type))))
                   (continue #t)]
                  [(defop)
                   (let* ([args (cdr statement)]
                          [op (car (list-ref args 0 #f))]
                          [type (let1 t (car (list-ref args 1 #f))
                                  (cond [(string=? t "left") 'left]
                                        [(string=? t "right") 'right]
                                        [(string=? t "prefix") 'prefix]
                                        [else (lmn-error #`"bad operator type ,|t|")]))]
                          [priority (let1 p (car (list-ref args 2 #f))
                                      (if (string-every #[0-9] p)
                                          (string->number p)
                                          (lmn-error #`"priority must be a natural number.")))]
                          [arity (and-let* ([a (list-ref args 3 #f)]
                                            [a (and (pair? a) (car a))]
                                            [a (and (string-every #[0-9] a)
                                                    (string->number a))])
                                   a)])
                     (unless (and type op priority)
                       (lmn-error "too few argumetns for defop."))
                     (cond [(assoc-ref *ops* priority) =>
                            (^l (unless (eq? type (car l))
                                  (lmn-error #`"invalid operator type (expected: ,(car l))."))
                                (push! (cdr l)
                                       (cons* op (^(:rest x) (cons op x)) arity)))]
                           [else
                            (push! *ops*
                                   (list priority type
                                         (cons* op (^(:rest x) (cons op x)) arity)))])
                     (print #`"defined ,|op|."))
                   (continue #t)]
                  [(command)
                   (case (cadr statement)
                     [(quit)
                      (break #t)]
                     [(clear)
                      (set! *process* (make-process))
                      (set! *count* 1)]
                     [(reset)
                      (set! *process* (make-process))
                      (set! *count* 1)
                      (set! *rules* ())
                      (set! *system-rules* (default-system))
                      (set! *types* (default-env))
                      (set! *ops* (default-ops))]
                     [(trace)
                      (set! *trace* (not *trace*))
                      (print #`"set to ,|*trace*|.")]
                     [(debug)
                      (set! *debug* (not *debug*))
                      (print #`"set to ,|*debug*|.")]
                     [(interactive)
                      (set! *interactive* (not *interactive*))
                      (print #`"set to ,|*interactive*|.")])
                   (continue #t)])))
            (while (any (cute apply-rule! *process* <> *types*) *rules*)
              (while (any (cute apply-rule! *process* <> *types*) *system-rules*))
              (dump-trace))
            (print (process->string *process*))
            (when *interactive*
              (set! *process* (make-process))
              (set! *count* 1))))))))

(define (main args)
  (print "\nContext-sensitive Flat LMNtal\n")
  (repl))

;; + (sentinel)
